__all__ = ["Buttons", "Button", "TextBox", "Slider", "DropdownBox", "Text"]

__version__ = "0.9.0"
__version_info__ = tuple(map(int, __version__.split(".")))

#Import all items so they are all actually loaded
from .Base import Buttons

from .Button import Button
from .TextBox import TextBox
from .Slider import Slider
from .DropdownBox import DropdownBox
from .Text import Text



from .Base import Buttons
from .Slider import Slider

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame


class Text(Buttons):
    """
    A simple (multi-line) text object, with scrolling support.

    pos: (left, top) - The topleft position before scaling.
    size: (width, height) - The size before scaling.
    text: str - The text that will be rendered to the surface.
    style: "Square", "Round", int - Defines the radius of curvature of the buttons' corners.
    font_name: str - The name of the font that should be used for the Text.
    font_size: int - The size (in px) of the text.
    text_colour: (R, G, B) - The colour of the text in the Text object.
    text_align: The alignment of the text on the Button surface.
    text_offset: "auto", int, (x, y) - The offset the text should have from the sides of the Text object. Prevents the text from overlapping with borders, and touching the edges.
    scroll_bar: None, int, Slider - The type of scrollbar to be included. Default styles 1 and 2 are available.
    background: pygame.Surface, (R, G, B), None, function - The background of the button.
    border: ((R, G, B), width, offset), None - The border that appears around the TextBox.
    functions: dict - Contains functions that should be called when a specific event occurs. The values should either be {"Click": func,} to call a function without arguments, or {"Click": (func, arg1, arg2, ...)} to call a function with arguments. If the Button itself is to be passed in as an argument, that argument can be passed in as '*self*'. This argument will automatically replaced when the function is actually called.
                    - "Move": Called whenever the Text object is scrolled.
    groups: None, [___, ___] - A list of all groups to which a button is to be added.
    root: None, Button - The Button that is considered the 'root element' for this Button. Any function calls that need to include a 'self' Button, will include this root Button instead.
    independent: bool - Determines whether or not the button is allowed to set the input_lock, and is added to buttons.list_all. Mostly important for buttons which are part of another button.

    Inputs:
    *.value: str - Can be used synonymously with *.text.
    *.text: str - Allows the user to set a new value for the Text objects' displayed text.
    *.lines: tuple - Allows the user to set a new value for 'lines' (the text as it is split to fit properly accros the lines).
    *.write(value) - Appends text to self.text. Allows this button to be used as an output for e.g. the print() function.

    Outputs:
    *.value: str - Synonymous with *.text.
    *.text: str - The current text being rendered to the surface.
    *.lines: tuple - The current text being rendered to the surface, as it is split to prevent it from exceeding the Surface borders.
    """
    actions = ["Scroll", "LMB_down", "LMB_up", "Set_cursor_pos"]
    def __init__(self, pos, size,
                 text = "",
                 style = "Square",
                 font_name = pygame.font.get_default_font(),
                 font_size = 22,
                 text_colour = (0, 0, 0),
                 text_align = "topleft",
                 text_offset = "auto",
                 scroll_bar = None,
                 background = None,
                 border = None,
                 functions = {},
                 group = None,
                 root = None,
                 independent = False,
                 ):
        """
        Create a Text Button object. See help(type(self)) for more detailed information.
        """
        super().__init__(pos, size, font_name, font_size, group, root, independent)
        self.style = style
        self.text_colour = self.Verify_colour(text_colour)
        self.text_align = text_align

        self.bg = self.Verify_background(background)
        self.border = self.Verify_border(border)

        #Set the offset the text has from the sides of the text_box
        if type(text_offset) is int:
            self.text_offset = 2 * (text_offset,)
        elif type(text_offset) is not str:
            self.text_offset = self.Verify_iterable(text_offset, 2)
        elif text_offset.lower() == "auto":
            #The automatic offset is calculated as 0.25 * font_size + (border_width + border_offset if there is a border)
            #Offset is not 0 if no border is given, to be consistent with TextBox Buttons
            #It can of course still be 0 if the user sets text_offset = 0
            self.text_offset = 2 * (round(self.font_size / 4) + ((border[1] + border[2]) if self.border else 0),)

        if scroll_bar:
            self.scroll_bar = Make_scroll_bar(self, scroll_bar)
            self.children.append(self.scroll_bar)
        else:
            self.scroll_bar = None
        self.__scrolled = 0
        self.functions = functions
        self.text = text
        self.Build_lines()
        self.Draw(pygame.Surface((1, 1))) #Makes sure all attributes are set-up correctly


    def LMB_down(self, pos):
        if self.scroll_bar:
            self.scroll_bar.LMB_down(pos)
            if self.Buttons.input_claim: #If the slider contained the position, and now claimed the input, set self as the lock
                self.Set_lock()


    def LMB_up(self, pos):
        if self.scroll_bar:
            self.scroll_bar.LMB_up(pos)
            if self.Buttons.input_claim:
                self.Release_lock()


    def Set_cursor_pos(self, pos):
        if self.scroll_bar:
            self.scroll_bar.Set_cursor_pos(pos)


    def Scroll(self, value, pos):
        if not self.contains(pos): #If the mouse was not within the text box:
            return
        self.scrolled_px += self.Buttons.scroll_factor * value
        self.Claim_input()


    def Scale(self, scale, relative_scale = True, *, center = (0, 0), px_center = None):
        super().Scale(scale, self, relative_scale, center = center, px_center = px_center)


    def Move(self, offset, scale = False):
        super().Move(offset, self, scale)


    def Clear(self):
        self.scrolled = 0
        self.Release_lock()
        #Does not do anything to the contained text, since that is not created directly through user input.


    def Draw(self, screen, pos = None):
        """
        Draw the button to the screen.
        """
        self.scrolled #Update the scrolled position quickly, so that any .moved = True are set
        pos = pos or self.scaled(self.topleft)

        if self.updated:
            self.Build_lines()
            self.moved = True

            #Make the background surface
            self.bg_surface = self.Make_background_surface(self.bg)
            if self.border:
                self.Draw_border(self.bg_surface, *self.border)

            font_height = self.font.get_height()
            if self.px_height >= len(self.lines) * font_height:
                #If the text fully fits within the available space, calculate the vertical offset to get the right alignment
                vert_offset = self.AlignY(len(self.lines) * font_height, self.px_height, self.text_align).top
            else:
                #If the text requires scrolling, vertical alignment doesn't matter anymore (all vertical alignment is taken over by the scrolled value)
                vert_offset = 0

            #Build the surface containing ALL lines of text
            self.text_surface =  pygame.Surface((self.px_width, font_height * len(self.lines) + vert_offset), pygame.SRCALPHA)

            for line_nr, line in enumerate(self.lines):
                line_surf = self.font.render(line, True, self.text_colour)
                line_rect = line_surf.get_rect()
                line_rect.top = line_nr * font_height + vert_offset
                line_rect = self.AlignX(line_rect, self.px_width, self.text_align)
                self.text_surface.blit(line_surf, line_rect)

            self.updated = False

        if self.moved:
            #Blit the fully rendered text surface onto a limiter surface.
            text_limiter = pygame.Surface((self.px_width, self.true_height - 2 * self.scaled(self.text_offset[1])), pygame.SRCALPHA)
            text_limiter.blit(self.text_surface, (0, -self.scrolled_px))

            #Blit the text surface onto the actual background
            self.surface = self.bg_surface.copy()
            self.surface.blit(text_limiter, self.scaled(self.text_offset))

            if self.scroll_bar:
                self.scroll_bar.Draw(self.surface, tuple(round(i) for i in self.relative(self.scroll_bar.scaled(self.scroll_bar.topleft))))
            self.moved = False

        screen.blit(self.surface, pos)
        return


    def write(self, value):
        """
        Append value to self.text.
        Allows for a Text object to be used as an output "file" for e.g. print.
        """
        self.text += value


    @property
    def scrolled(self):
        if self.scroll_bar and self.scroll_bar.moved:
            self.__scrolled = self.scroll_bar.value
            self.moved = True
        return self.__scrolled

    @scrolled.setter
    def scrolled(self, value):
        #Make sure the scrolled value cannot exceed the limits of the space in the box
        value = self.Clamp(value, 0, 1)
        self.moved = True

        self.__scrolled = value
        if self.scroll_bar:
            self.scroll_bar.value = value

        return

    @property
    def scrolled_px(self):
        #Calculate the required height (in px)
        req_height = self.font.get_height() * len(self.lines)
        #Calculate the required height change that has to be accomodated by scrolling (in px)
        max_scroll_height = max(0, req_height - self.true_height + 2 * self.scaled(self.text_offset[1])) #Get the total distance (in px) the surface must be scrolled

        return round(max_scroll_height * self.scrolled)

    @scrolled_px.setter
    def scrolled_px(self, value):
        #Calculate the required height (in px)
        req_height = self.font.get_height() * len(self.lines)
        #Calculate the required height change that has to be accomodated by scrolling (in px)
        max_scroll_height = max(0, req_height - self.true_height + 2 * self.scaled(self.text_offset[1])) #Get the total distance (in px) the surface must be scrolled

        if max_scroll_height:
            self.scrolled = value / max_scroll_height


    @property
    def value(self):
        return self.text

    @value.setter
    def value(self, val):
        self.text = val


    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, value):
        if type(value) is not str:
            raise TypeError(f"Text should be type str, not type {type(value).__name__}.")

        self.__text = value
        self.updated = True


    @property
    def lines(self):
        return self.__lines

    @lines.setter
    def lines(self, value):
        """

        """
        #For external use only. Internally, all writing calls are directly to self.__lines
        if type(value) not in (tuple, list,):
            raise TypeError(f"Lines must be type 'tuple' or type 'list', not type {type(value).__name__}")
        self.__lines = tuple(value)
        self.__text = "\n".join(self.__lines)
        self.updated = True


    def Build_lines(self):
        """
        (Re-)builds the '*.lines' tuple based on the current value of self.text, such that the text will automatically wrap around to the next line if it won't fit on the current line anymore.
        Called automatically in *.Draw, after *.text is set / changed.
        """
        max_width = self.px_width
        #Split the text into lines, ignoring any trailing newlines.
        text_lines = self.text.rstrip("\n").split("\n")
        lines = []
        for line in text_lines:
            words = line.split(" ")
            line_string = words[0]
            for word in words[1:]:
                if self.font.size(" ".join([line_string, word]))[0] <= max_width: #If the next word still fits on this line:
                    line_string = " ".join([line_string, word]) #Join it together with the existing text
                else: #If the word is too long to fit on the line:
                    lines.append(line_string)
                    line_string = word #Place it on the next line.
            #Once all words are exhausted, append the remaining string to lines as well
            lines.append(line_string)
        self.__lines = tuple(lines)

        if self.scroll_bar:
            self.scroll_bar.Set_slider_primary(round(self.scroll_bar.height * min(1, (self.height - 2 * self.text_offset[1]) / (len(self.lines) * self.font_size))))

        self.scrolled += 0 #Update the 'scrolled' value, to take into account that after rebuilding, the length of 'lines' might be different

    @property
    def px_width(self):
        """
        The maximum width (in px) the text may have. In other words, the horizontal space available for text.
        """
        return self.true_width - 2 * self.scaled(self.text_offset[0]) - (self.scroll_bar.true_width + self.scaled(self.text_offset[0]) if self.scroll_bar else 0)

    @property
    def px_height(self):
        """
        The maximum height (in px) the text may have. In other words, the horizontal space available for text.
        """
        return self.true_height + 2 * self.scaled(self.text_offset[1])

def Make_scroll_bar(self, scroll_bar):
    """
    Make a scroll_bar for a Text object.
    For internal use only. This function is therefore also not imported by __init__.py
    """
    if isinstance(scroll_bar, Slider):
        scroll_bar.right = self.right - self.text_offset[0]
        scroll_bar.top = self.top + self.text_offset[1]
        if scroll_bar.height > self.height - 2 * self.text_offset[1]:
            scroll_bar.height = self.height - 2 * self.text_offset[1]
        return scroll_bar
    if scroll_bar == 1:
        size = (15, self.height - 2 * self.text_offset[1])
        pos = (self.right - size[0] - self.text_offset[0], self.top + self.text_offset[1])
        style = "Round"
        background = None
        border = None
        slider_bg = (220, 220, 220)
        slider_accent_bg = (127, 127, 127)
        slider_border = None
        return Slider(pos, size, style = style, background = background, border = border, slider_background = slider_bg, slider_border = slider_border, root = self.root, independent = True)
    elif scroll_bar == 2:
        size = (15, self.height - 2 * self.text_offset[1])
        pos = (self.right - size[0] - self.text_offset[0], self.top + self.text_offset[1])
        slider_feature_text = "|||"
        slider_feature_size = 9
        return Slider(pos, size, slider_feature_text = slider_feature_text, slider_feature_size = slider_feature_size, root = self.root, independent = True)
    else:
        raise ValueError(f"Unsupported scroll_bar style: {repr(scroll_bar)}")



from .Base import Buttons

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame


class TextBox(Buttons):
    """
    Creates a TextBox, in which a user can input text.

    pos: (left, top) - The topleft position before scaling.
    size: (width, height) - The size before scaling.
    hint: str - The text that will be shown if no text is input by the user.
    style: "Square", "Round", int - Defines the radius of curvature of the buttons' corners.
    font_name: str - The name of the font that should be used for the TextBox.
    font_size: int - The size (in px) of the text.
    text_colour: (R, G, B) - The colour of the text the user types.
    hint_colour: (R, G, B) - The colour of the hint.
    text_align: The alignment of the text on the Button. Vertical alignment is always active. Horizontal alignment only if the width of the text is less than the available space.
    text_offset: "auto", int, (x, y) - The offset the text should have from the sides of the TextBox. Prevents the text from overlapping with borders, and touching the edges.
    background: pygame.Surface, (R, G, B), None, function - The background of the button if it is not selected.
    border: ((R, G, B), width, offset), None - The border that appears around the TextBox.
    accent background: pygame.Surface, (R, G, B), None, function - The background of the button if it is_selected. If set to None, will be the same as normal background.
    accent_border: ((R, G, B), width, offset), None - An additional border that can be drawn when the TextBox is selected.
    functions: dict - Contains functions that should be called when a specific event occurs. The values should either be {"Click": func,} to call a function without arguments, or {"Click": (func, arg1, arg2, ...)} to call a function with arguments. If the Button itself is to be passed in as an argument, that argument can be passed in as '*self*'. This argument will automatically replaced when the function is actually called.
                    - "Select": Called whenever the TextBox is selected.
                    - "Deselect": Called whenever the TextBox is deselected.
                    - "Type": Called every time a valid Key_down (one which could alter the contents of the TextBox) is recorded while this TextBox is selected.
    groups: None, [___, ___] - A list of all groups to which a button is to be added.
    root: None, Button - The Button that is considered the 'root element' for this Button. Any function calls that need to include a 'self' Button, will include this root Button instead.
    independent: bool - Determines whether or not the button is allowed to set the input_lock, and is added to buttons.list_all. Mostly important for buttons which are part of another button.

    Inputs:
    *.value: str - Sets the current text in the TextBox.
    *.text: str - Synonymous to *.value. Can be used to keep code clearer / more readable, depending on the context of where this button is used.

    Outputs:
    *.value: str - The current value in the TextBox. I.E. the text input by the user into the input field.
    *.text: str - Synonymous to *.value. Can be used instead to keep code clearer / more readable, depending on the context of where this button is used.
    *.new_input: bool - Whether the TextBox has received any new text inputs since the last time this variable was checked. Automatically resets once it is querried.
    *.deselected: bool - Whether the TextBox was deselected since the last time this variable was checked. Automatically resets once it is querried.

    *.is_selected: bool - Whether this TextBox object is selected at this point in time. I.E. Whether the user is currently typing in this TextBox.
    """
    actions = ["LMB_down", "Key_down"]
    def __init__(self, pos, size,
                 hint = "",
                 style = "Square",
                 font_name = pygame.font.get_default_font(),
                 font_size = 22,
                 text_colour = (0, 0, 0),
                 hint_colour = (128, 128, 128),
                 text_align = "left",
                 text_offset = "auto",
                 background = (255, 255, 255),
                 border = ((63, 63, 63), 1, 0),
                 accent_background = None,
                 accent_border = ((0, 0, 0), 1, 2), #Set to None or False to disable
                 functions = {},
                 group = None,
                 root = None,
                 independent = False,
                 ):
        """
        Create a TextBox Button object. See help(type(self)) for more detailed information.
        """
        super().__init__(pos, size, font_name, font_size, group, root, independent)
        #Set up of basic TextBox properties
        self.text = ""
        self.new_input = False
        self.hint = hint
        self.style = style
        self.text_colour = self.Verify_colour(text_colour)
        self.hint_colour = self.Verify_colour(hint_colour)
        self.text_align = text_align
        self.bg = self.Verify_background(background)
        if accent_background:
            self.accent_bg = self.Verify_background(accent_background)
        else:
            self.accent_bg = self.bg

        #Verify and set the border variables
        self.border = self.Verify_border(border)
        self.accent_border = self.Verify_border(accent_border)

        #Set the offset the text has from the sides of the text_box. In the end,
        #text_offset should be a tuple (x_offset, y_offset)
        if type(text_offset) is int:
            self.text_offset = 2 * (text_offset,)
        elif type(text_offset) is not str:
            self.text_offset = self.Verify_iterable(text_offset, 2)
        elif text_offset.lower() == "auto":
            #The automatic offset is calculated as 0.25 * font_size + max(border_width + border_offset for any of the borders)
            self.text_offset = 2 * (round(self.font_size / 4) + max([brdr[1] + brdr[2] for brdr in (self.border, self.accent_border) if brdr], default = 0),)

        #Settting the initial state for certain default variables
        self.text_scroll = 0
        self.cursor = 0
        self.__is_selected = False
        self.deselected = False
        self.functions = functions
        self.Draw(pygame.Surface((1, 1))) #Makes sure all attributes are set-up correctly


    def LMB_down(self, pos):
        if self.contains(pos):
            self.Claim_input()
            if self.is_selected:
                pos = self.relative(pos)
                #If there is any text: (Check required since for loop has to run at least once to not crash)
                if self.text:
                    #Iterate over all letters, to find which letter was closest to
                    #the position at which the user clicked
                    pixel_offset = - self.text_scroll + self.scaled(max(self.border[1] + self.border[2] + round(self.font_size / 4), self.accent_border[1] + self.accent_border[2] + round(self.font_size / 4)), True)
                    for letter_nr, letter in enumerate(self.text):
                        pixel_length = self.font.size(self.text[:letter_nr + 1])[0]
                        if (pixel_length + pixel_offset) >= pos[0]:
                            break
                    #Calculate the horizontal distance from the cursor to the text box
                    distance = pixel_length + pixel_offset - pos[0]
                    #Get the size of the last letter in the list
                    letter_size = self.font.size(letter)[0]
                    #If the cursor is more than halfway back before the end of this letter
                    #put the cursor in front of the letter.
                    if distance >= (0.5 * letter_size):
                        self.cursor = letter_nr
                    #Else, put it after the letter
                    else:
                        self.cursor = letter_nr + 1
                #If there is no text:
                else:
                    self.cursor = 0
            else:
                self.is_selected = True
                self.cursor = len(self.text)
        elif self.is_selected:
            self.is_selected = False
            self.Buttons.input_processed = True

        return


    def Key_down(self, event):
        if event.key in (pygame.K_RETURN, pygame.K_ESCAPE):
            self.is_selected = False
        elif event.key == pygame.K_BACKSPACE:
            self.text = self.text[:max(self.cursor - 1, 0)] + self.text[self.cursor:]
            #Move the cursor back one item
            self.cursor -= 1
        elif event.key == pygame.K_DELETE:
            self.text = self.text[:self.cursor] + self.text[self.cursor + 1:]
        elif event.key == pygame.K_LEFT:
            self.cursor -= 1
        elif event.key == pygame.K_RIGHT:
            self.cursor += 1
        elif event.unicode:
            self.text = self.text[:self.cursor] + event.unicode + self.text[self.cursor:]
            #Scroll the item sideways
            #self.text_scroll += self.font.size(event.unicode)[0]
            self.cursor += 1
        else:
            return
        #Inform Buttons that the input has been processed / used
        self.Buttons.input_processed = True
        #Call "type" if the Key_down event was one which could have altered the text contents.
        if event.key not in (pygame.K_RETURN, pygame.K_RIGHT, pygame.K_LEFT):
            self.root._Call("Type")
        return


    def Scale(self, scale, relative_scale = True, *, center = (0, 0), px_center = None):
        super().Scale(scale, self, relative_scale, center = center, px_center = px_center)


    def Move(self, offset, scale = False):
        super().Move(offset, self, scale)


    def Clear(self):
        self.value = ""
        self.is_selected = False
        #Lock is automatically released in property setter


    def Draw(self, screen, pos = None):
        """
        Draw the button to the screen.
        """
        pos = pos or self.scaled(self.topleft)
        if self.updated:
            self.update_scroll()
            #Draw the correct background onto the surface
            if not self.is_selected:
                self.surface = self.Make_background_surface(self.bg)
            else:
                self.surface = self.Make_background_surface(self.accent_bg)
            #Draw a border, if it is enabled
            if self.border:
                self.Draw_border(self.surface, *self.border)
            #Draw a accent border, if it is enabled:
            if self.accent_border and self.is_selected:
                self.Draw_border(self.surface, *self.accent_border)
            #Copy the surface to allow the cursor to be drawn
            self.cursor_surface = self.surface.copy()

            #Add the text to the surface
            text_limiter = pygame.Surface(self.offset(self.true_size, self.scaled(self.text_offset), (-2, -2)), pygame.SRCALPHA)
            limiter_rect = text_limiter.get_rect()
            if self.text:
                text_surface = self.font.render(self.text, True, self.text_colour)
            else:
                text_surface = self.font.render(self.hint, True, self.hint_colour)
            text_rect = text_surface.get_rect()
            #Align the text rect
            text_rect = self.AlignY(text_rect, limiter_rect, self.text_align)
            if text_rect.width <= limiter_rect.width:
                #If the text is smaller than the limiter, perform alignment in the X direction
                self.AlignX(text_rect, limiter_rect, self.text_align)
            else:
                #If the text is wider than the limiter, all alignment is taken care of inside text_scroll
                text_rect.left = - self.text_scroll
            text_limiter.blit(text_surface, text_rect)
            #Blit the limiter onto the button
            limiter_rect.center = self.middle #middle is scaled(width / 2, height / 2)
            self.surface.blit(text_limiter, limiter_rect)

            #Make the cursor surface
            cursor_rect = pygame.Rect((0, 0), (max(1, self.scaled(1)), self.font.get_height()))
            #Align cursor vertically
            cursor_rect.centery = text_rect.centery
            #Align cursor horizontally. (if-else statement is required to prevent the hint from changing the Cursor location.)
            cursor_rect.left = self.font.size(self.text[:self.cursor])[0] + (text_rect.left if self.text else self.AlignX(cursor_rect.width, limiter_rect, self.text_align).left)
            #Draw the cursor to the text limiter
            pygame.draw.rect(text_limiter, self.text_colour,  cursor_rect)
            self.cursor_surface.blit(text_limiter, limiter_rect)

            #Clear self.updated again, as the surface has been remade.
            self.updated = False

        if self.is_selected:
            #Update the cursor animation
            self.cursor_animation = (self.cursor_animation + 1) % self.framerate
        if self.cursor_animation < round(self.framerate / 2):
            screen.blit(self.cursor_surface, pos)
        else:
            screen.blit(self.surface, pos)
        return


    @property
    def is_selected(self):
        return self.__is_selected

    @is_selected.setter
    def is_selected(self, value):
        #If the user selects the text box:
        if value:
            self.__is_selected = True
            self.cursor = len(self.text)
            self.Set_lock()
            self.root._Call("Select")
        else:
            self.__is_selected = False
            self.deselected = True
            self.cursor = 0
            self.cursor_animation = self.framerate
            self.Release_lock(False) #Release without claiming the input
            self.root._Call("Deselect")


    @property
    def deselected(self):
        deselected_ = self.__deselected
        self.__deselected = False
        return deselected_

    @deselected.setter
    def deselected(self, value):
        self.__deselected = value


    @property
    def cursor(self):
        return self.__cursor

    @cursor.setter
    def cursor(self, value):
        #Make sure the cursor cannot be set to negative points, nor can it go further than directly after the last character.
        self.__cursor = self.Clamp(int(value), 0, len(self.text))
        self.cursor_animation = self.framerate - 1
        self.updated = True
        self.update_scroll()


    @property
    def text_scroll(self):
        return self.__text_scroll

    @text_scroll.setter
    def text_scroll(self, value):
        self.__text_scroll = value
        self.updated = True


    @property
    def text(self):
        return self.__value

    @text.setter
    def text(self, value):
        self.__value = value
        self.updated = True
        self.new_input = True


    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, val):
        self.__value = val
        self.updated = True
        self.new_input = True


    @property
    def new_input(self):
        new_input_ = self.__new_input
        self.__new_input = False
        return new_input_

    @new_input.setter
    def new_input(self, value):
        self.__new_input = value


    def update_scroll(self):
        """
        Update the value of the scrolling (within limits).
        """
        #Get the width of the text box; +1 to account for a possible cursor at the end.
        #Always add this +1, to prevent annoying 1-pixel shifts when moving the cursor to the final position.
        text_width = self.font.size(self.text)[0] + 1
        #Get the width of the text limiter surface
        limiter_width = self.true_width - self.scaled(2 * self.text_offset[0])
        #Get the cursor pixel index; +1 not required since 'size' already includes index 0 as width 1
        cursor_pos = self.font.size(self.text[:self.cursor])[0]
        #If all text fits in the view window:
        if text_width <= limiter_width:
            #Reset any scroll. No need to scroll if it fits anyway
            self.text_scroll = 0
        #If the text is bigger than the scroll window
        else:
            self.text_scroll = self.Clamp(self.text_scroll, 0, text_width - limiter_width)
            #If the cursor is before the view window:
            if cursor_pos < self.text_scroll:
                self.text_scroll = cursor_pos
            #If the cursor is after the view window:
                # => because the cursor is 1 pixel wide, and thus ends at pos+1
            elif cursor_pos >= self.text_scroll + limiter_width:
                self.text_scroll = cursor_pos - limiter_width + 1
            #If the cursor is already inside of the view window:
            else:
                pass



from .Base import Buttons

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame
import math


class Button(Buttons):
    """
    Creates a Button, an object which can detect Left Mouse Button inputs from the user.

    pos: (left, top) - The topleft position before scaling.
    size: (width, height) - The size before scaling.
    text: str -
    mode: str - "None": Does not change *.value when clicked. Will set *.clicked and execute any assigned functions.
              - "Count": Count the number of inputs that have occured.
              - "Toggle": Toggle value from True to False and back.
              - "Hold": Stays True until the Mouse Button is released.
    orientation: int - The orientation of the text. If orientation == 0, the text will be horizontal; if orientation == 1, the text will be vertical.
    style: "Square", "Round", int - Defines the radius of curvature of the buttons' corners.
    font_name: str - The name of the font that should be used for the Button.
    font_size: int - The size (in px) of the text.
    text_colour: (R, G, B) - The colour of the text the user types.
    text_align: The alignment of the text on the Button.
    background: pygame.Surface, (R, G, B), None, function - The background of the button if it is not selected.
    border: ((R, G, B), width, offset), None - The border that appears around the TextBox.
    accent background: pygame.Surface, (R, G, B), None, function - The background of the Button if *.value. If set to None, will be the same as normal background.
    dragable: (horizotal, vertical) - A tuple of two booleans defining whether the Button is allowed to be moved in either the horizontal and / or vertical direction repectively. Requires mode == "Hold".
    limits: (left, right, top, bottom) - The coordinate limits between which the Button is to be draggable.
    snap: ((x snap coords, ___), (y snapc coords, ___), snap_range) - If the button is dragable / movable, the positions to which the Button should snap, as well as the range (in px) within which the Button should snap to these locations.
    functions: dict - Contains functions that should be called when a specific event occurs. The values should either be {"Click": func,} to call a function without arguments, or {"Click": (func, arg1, arg2, ...)} to call a function with arguments. If the Button itself is to be passed in as an argument, that argument can be passed in as '*self*'. This argument will automatically replaced when the function is actually called.
                    - "Click": Called when the Button is clicked.
                    - "Release": Called when the Button is released. Available only when mode == "Hold" or mode == "Toggle".
                    - "Move": Called when the Button is dragged to a new location. Only available if any(dragable).
    groups: None, [___, ___] - A list of all groups to which a button is to be added.
    root: None, Button - The Button that is considered the 'root element' for this Button. Any function calls that need to include a 'self' Button, will include this root Button instead.
    independent: bool - Determines whether or not the button is allowed to set the input_lock, and is added to buttons.list_all. Mostly important for buttons which are part of another button.

    Inputs:
    *.value: int, bool - The current state of the button.

    Outputs:
    *.value: int, bool - The current state in the Button. If mode == "count", the amount of times (int) the Button was clicked. If mode == "Toggle" or "hold", whether the button is currently in the pressed / down state (bool).
    *.clicked: bool - Whether the Button has been set to a new state since the last time this variable was checked. Automatically resets once it is querried.
    """
    actions = ["LMB_down", "LMB_up", "RMB_down", "Set_cursor_pos"]
    def __init__(self, pos, size,
                 text = "",
                 mode = "None",
                 orientation = 0,
                 style = "Square",
                 font_name = pygame.font.get_default_font(),
                 font_size = 22,
                 text_colour = (0, 0, 0),
                 text_align = "Center",
                 background = (255, 255, 255),
                 border = ((63, 63, 63), 1, 0),
                 accent_background = (220, 220, 220),
                 dragable = (False, False),
                 limits = (None, None, None, None),
                 snap = ((), (), 0),
                 functions = {},
                 group = None,
                 root = None,
                 independent = False
                 ):
        """
        Create a Button Button object. See help(type(self)) for more detailed information.
        """
        super().__init__(pos, size, font_name, font_size, group, root, independent)
        self.orientation = orientation
        self.style = style
        if type(mode) is not str:
            raise TypeError(f"mode must be type 'str', not type {type(mode).__name__}")
        elif mode.lower() == "count":
            self.value = 0
        elif mode.lower() in ("none", "toggle", "hold"):
            self.value = False
        else:
            raise ValueError(f"Unsupported button mode '{mode}'")
        self.mode = mode.lower()
        self.clicked = False #Make sure the button does not register any ghost inputs at the start
        self.text = text
        self.text_colour = self.Verify_colour(text_colour)
        self.text_align = text_align
        self.bg = self.Verify_background(background)
        if accent_background:
            self.accent_bg = self.Verify_background(accent_background)
        else:
            self.accent_bg = self.bg
        self.border = self.Verify_border(border)
        self.dragable = self.Verify_iterable(dragable, 2, bool)
        limits = self.Verify_iterable(limits, 4)
        self.limits = list(value if value is not None else ( (-1) ** (i + 1) * math.inf) for i, value in enumerate(limits))
        self.snap = self.Verify_iterable(snap, 3)
        self.functions = functions
        self.Draw(pygame.Surface((1, 1))) #Makes sure all attributes are prepared and set-up correctly


    def LMB_down(self, pos):
        if self.contains(pos):
            self.clicked = True
            if self.mode == "none": #Just here for consistency / clarity. Does nothing.
                pass
            elif self.mode == "count":
                self.value += 1
            elif self.mode == "toggle":
                self.value = not self.value
            elif self.mode == "hold":
                self.Set_lock()
                self.value = True
                if any(self.dragable):
                    self.drag_pos = self.relative(pos)
            self.Claim_input()

            if self.mode == "toggle" and not self.value: #If the button is toggled OFF, _Call release. For all other cases, _Call click
                self.root._Call("Release")
            else:
                self.root._Call("Click")
        return

    def LMB_up(self, pos):
        if any(self.dragable):
            self.Set_cursor_pos(pos)
        if self.mode == "hold":
            if self.value:
                self.value = False
                self.clicked = True
                self.Release_lock()
                self.root._Call("Release")
        return

    def Set_cursor_pos(self, pos):
        if self.value:
            if any(self.dragable):
                topleft = self.topleft
                hori, verti = self.offset(self.relative(pos), self.drag_pos, (-1, -1)) #Determine the offsets of the cursor from where the user originally clicked

                left = self.left + hori / self.scale * self.dragable[0] #Scale the offsets down to the original scale
                top = self.top + verti / self.scale * self.dragable[1]
                #Perform snapping
                distances = sorted(zip((abs(left - snap_point) for snap_point in self.snap[0]), self.snap[0]))
                if distances: #If there are any snap point:
                    closest = distances[0]
                    if closest[0] <= self.snap[2]: #If the distance <= the snapping range
                        left = round(closest[1]) #Snap!
                #Also, vertical snapping
                distances = sorted(zip((abs(top - snap_point) for snap_point in self.snap[1]), self.snap[1]))
                if distances: #If there are any snap point:
                    closest = distances[0]
                    if closest[0] <= self.snap[2]: #If the distance <= the snapping range
                        top = round(closest[1]) #Snap!
                #Confine the button within the limits
                self.left = self.Clamp(left, self.limits[0], self.limits[1] - self.width)
                self.top = self.Clamp(top, self.limits[2], self.limits[3] - self.height)
                if self.topleft != topleft: #If the Button moved:
                    self.root._Call("Move")


    def Scale(self, scale, relative_scale = True, *, center = (0, 0), px_center = None):
        super().Scale(scale, self, relative_scale, center = center, px_center = px_center)


    def Move(self, offset, scale = False):
        super().Move(offset, self, scale)


    def Clear(self):
        self.value = 0 if self.mode.lower() == "count" else False
        self.Release_lock()


    def Draw(self, screen, pos = None):
        """
        Draw the button to the screen.
        """
        if self.updated:
            #Build the correct background for the surface
            if not self.value:
                self.surface = self.Make_background_surface(self.bg)
            else:
                self.surface = self.Make_background_surface(self.accent_bg)
            #Draw a border, if it is enabled
            if self.border:
                self.Draw_border(self.surface, *self.border)
            #Draw the text onto the surface
            if self.text:
                #Make a surface that fits within the border
                text_offset = self.scaled(self.border[1] + self.border[2] + self.font_size / 4 if self.border else self.font_size / 4)
                text_limiter = pygame.Surface(self.Clamp(self.offset(self.true_size, 2 * (text_offset,), (-2, -2)), 0, math.inf), pygame.SRCALPHA)
                limiter_rect = text_limiter.get_rect()
                text_surface = self.font.render(self.text, True, self.text_colour)

                if self.orientation:
                    text_surface = pygame.transform.rotate(text_surface, -90 * self.orientation)
                text_rect = text_surface.get_rect()
                #Align the text properly
                self.Align(text_rect, limiter_rect, self.text_align)
                #Blit the text to the limiter surface, and then onto the screen
                text_limiter.blit(text_surface, text_rect)
                limiter_rect.center = self.middle
                self.surface.blit(text_limiter, limiter_rect)

            #Clear self.updated again, as the surface has been remade.
            self.updated = False
        screen.blit(self.surface, pos or self.scaled(self.topleft))
        return

    def _move(self, value):
        self.left += value[0]
        self.top += value[1]
        self.limits[0] += value[0]
        self.limits[1] += value[0]
        self.limits[2] += value[1]
        self.limits[3] += value[1]



    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, value):
        self.__text = str(value)
        self.updated = True

    @property
    def text_align(self):
        return self.__text_align
    @text_align.setter
    def text_align(self, value):
        self.__text_align = value
        self.updated = True

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value
        self.updated = True

    @property
    def clicked(self):
        clicked_ = self.__clicked
        self.__clicked = False
        return clicked_

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value



import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame

import math

pygame.font.init() #Required to set a font



class Buttons():
    """
    A class to serve as a base for all Button classes. Also provides interfaces for button-to-button "communication" (e.g. claiming a cursor lock)
    Furthermore, provides certain functionalities that make it easier to implement buttons into a program. (E.G. a way to Click all relevant buttons)

    Capitalisation:
    All actions that can be applied to a button (e.g. LMB_down) are capitalised.
    All attributes and functions that do not directly alter anything about the button itself (e.g. scaling a variable to the Buttons' scale) are not capitalised.

    Available Actions - for more detailed information, see help(Buttons.*function_name*):
    Buttons.Event(pygame.Event, group) - Allows for a pygame.event to be processed completely autonomously, and be diverted to the correct buttons. No further intervention is required if this function is used.
    Buttons.Update(group) - Updates all Buttons in the given group wherever necessary, such as updating the current cursor position for active sliders. Should be called once at the end of each input / event cycle.
    Buttons.Draw(screen, group) - Draw all Buttons in the given group(s) to the given screen / pygame.Surface
    Buttons.Scale(scale, group) - Scales all Buttons in the given group to / by the given factor.
    Buttons.Move(offset, group) - Moves all Buttons in the given group by the given offset.
    Buttons.Clear(group) - Clears all user inputs from Buttons. Note: Does NOT remove the text from Text objects.

    Other Actions (automatically called by Buttons.Event() / Buttons.Update() when required):
    Buttons.LMB_down(pos, group) - Perform a LMB_down (normal mouse click) at a certain position.
    Buttons.LMB_up(pos, group) - Perform a LMB_up (releasing a normal mouse click) at a certain position.
    Buttons.Scroll(value, pos, group) - Perform a Scrolling action at a certain position.
    Buttons.Key_down(event, group) - Perform a Key_down (typing) action.
    Buttons.Set_cursor_pos(pos, group) - Update the cursor position.

    Available functions:
    Buttons.get_group(group) - Will return a list of all Buttons that are in the given group(s). Also used internally when getting all buttons for which an Action should be applied.

    Class attributes:
    Buttons.input_claim - Contains whether or not the last input was fully claimed by a Button. E.G. If a DropdownBox was extended by clicking on  the Arrow button.
    Buttons.input_processed - Contains whether or not the last input was handled by a Button, even if they did not fully claim it. E.G. when exiting a TextBox by clicking outside of the TextBox area.


    Individual Button functions:
    *.get_rect() - Get a pygame.Rect object for the Button.
    *.get_scaled_rect() - Get a pygame.Rect object for the Button at its current scale.
    *.Add_to_group(group) - Add the Button on which this is called to the given group(s).
    """
    #A base class for all buttons
    input_lock = None #Either None, or the currently selected button. Used to give the currently selected button input priority.
    input_claim = False #Set to True if a button has claimed the input, to prevent an input from affecting multiple buttons.
    input_processed = False
    list_all = [] #A list containing all buttons, except those marked as independent. Can be used for debugging, or just to keep a nice list of all buttons.
    groups = {} #Groups to be used for getting certain buttons.
    scroll_factor = 1 #A factor to multiply scrolling with. Should be set based on the target DPI / resolution of the program
    #A framerate variable to help with timing animations
    framerate = 30
    min_scale = 0.05
    max_scale = 5

    def __init__(self, pos, size, font_name = pygame.font.get_default_font(), font_size = 22, groups = None, root = None, independent = False):
        #Tasks that are the same for all sub-classes
        self.updated = True
        self.children = []
        self.scale = 1

        #Tasks that require information from the child class
        self.size = size
        self.topleft = pos
        #Font size has to be set before font name, as setting font name prompts the font object to be built
        self.font_size = font_size
        self.font_name = font_name
        self.Add_to_group(groups)
        self.root = self if root is None else root
        if not independent:
            self.Buttons.list_all.append(self)
        self.independent = independent


    def __str__(self):
        return f"{type(self).__name__} object"
    def __repr__(self):
        return f"<{self.__str__()} at {self.topleft}>"



    @classmethod
    def get_group(cls, group):
        """
        Returns all buttons inside the given group / groups.
        """
        #If a list of groups is given, return the buttons in all the groups combined.
        if type(group) in (tuple, list):
            lst = []
            for grp in group:
                if grp in cls.groups:
                    for button in cls.groups[grp]:
                        #Append all buttons in the group to the original group, if it is not a duplicate
                        if button not in lst:
                            lst.append(button)
                #Else, if the group is already a Button, append that Button to the list instead
                elif isinstance(grp, Buttons):
                    lst.append(grp)
            return lst
        #Select the correct button group
        elif group is all: #If the group is the default 'all', return all buttons
            return cls.list_all
        elif group in cls.groups:
            return cls.groups[group] #Return all buttons in the group.
        elif isinstance(group, Buttons):
            return [group]
        else: #If the group doesn't exist, return an empty list
            return []

    def Add_to_group(self, groups):
        """
        Add a button to a group.
        Allows for assignment of multiple groups simultaniously by passing in a list or tuple of groups.
        """
        if not isinstance(groups, (list, tuple)):
            groups = [groups]
        for grp in groups:
            if grp is None:
                continue
            if grp in self.Buttons.groups:
                #If the group exists, add self to the group, unless self is already in this group.
                if not self in self.Buttons.groups[grp]:
                    self.Buttons.groups[grp].append(self)
            #If the group doesn't exist, make a new group with self as the first list entry.
            else:
                self.Buttons.groups[grp] = [self]

    def Set_lock(self, claim = True):
        """
        Set the input lock (if possible).
        If claim = True, automatically set Buttons.input_claim as well.
        """
        self.Buttons.input_processed = True
        if not self.Buttons.input_lock and not self.independent:
            self.Buttons.input_lock = self
        if claim:
            self.Buttons.input_claim = True

    def Release_lock(self, claim = True):
        """
        Release the input lock (if necessary / possible).
        If claim = True, automatically set Buttons.input_claim as well.
        """
        self.Buttons.input_processed = False
        if self is self.Buttons.input_lock and not self.independent:
            self.Buttons.input_lock = None
        if claim:
            self.Buttons.input_claim = True

    @classmethod
    def Claim_input(cls):
        cls.Buttons.input_claim = True
        cls.Buttons.input_processed = True


    @classmethod
    def Event(cls, event, group = all):
        """
        A method to handle all events a button might need to be informed of.
        """
        #Reset the input_claim and input_processed attributes
        cls.input_claim = False
        cls.input_processed = False

        #Handle the Event appropriately
        if type(event) != pygame.event.EventType:
            raise TypeError(f"Event should be type 'Event', not type {type(event).__name__}")
        if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 1:
                cls.LMB_down(event.pos, group)
            elif event.button == 2:
                cls.RMB_down(event.pos, group)
            elif event.button > 3:
                cls.Scroll(cls.convert_scroll(event.button), event.pos, group)
        elif event.type == pygame.MOUSEBUTTONUP:
            if event.button == 1:
                cls.LMB_up(event.pos, group)
            elif event.button == 2:
                cls.RMB_up(event.pos, group)
        elif event.type == pygame.KEYDOWN:
            cls.Key_down(event, group)


    @classmethod
    def Update(cls, group = all):
        """
        A method that will run any updates that have to be performed each cycle, such as updating the cursor position for sliders.
        """
        if cls.input_lock:
            if "Set_cursor_pos" in cls.input_lock.actions:
                cls.Set_cursor_pos(pygame.mouse.get_pos(), group)


    @classmethod
    def LMB_down(cls, pos, group = all):
        """
        Left Mouse Button down; A.K.A. a normal click.
        """
        cls.input_claim = False
        cls.input_processed = False
        group_list = cls.get_group(group)

        #If a button has claimed an input lock
        if cls.input_lock in group_list:
            if "LMB_down" in cls.input_lock.actions:
                cls.input_lock.LMB_down(pos)
                if cls.input_claim:
                    return

        for button in group_list:
            if "LMB_down" in button.actions and button is not cls.input_lock:
                button.LMB_down(pos)
                if cls.input_claim:
                    return
        return


    @classmethod
    def LMB_up(cls, pos, group = all):
        """
        Left Mouse Button up; A.K.A. releasing a click.
        """
        cls.input_claim = False
        cls.input_processed = False
        group_list = cls.get_group(group)

        #If a button has claimed an input lock
        if cls.input_lock in group_list:
            if "LMB_up" in cls.input_lock.actions:
                cls.input_lock.LMB_up(pos)
                if cls.input_claim:
                    return

        for button in group_list:
            if "LMB_up" in button.actions and button is not cls.input_lock:
                button.LMB_up(pos)
                if cls.input_claim:
                    return
        return


    @classmethod
    def Scroll(cls, value, pos, group):
        """
        Handles all scroll events for all buttons.
        """
        cls.input_claim = False
        cls.input_processed = False
        group_list = cls.get_group(group)

        if cls.input_lock in group_list:
            if "Scroll" in cls.input_lock.actions:
                cls.input_lock.Scroll(value, pos)
                if cls.input_claim:
                    return

        for button in group_list:
            #If the button hasn't been processed yet in the input_lock section, and has the "Scroll" attribute:
            if "Scroll" in button.actions and button is not cls.input_lock:
                button.Scroll(value, pos)
                if cls.input_claim:
                    return
        return


    @classmethod
    def Key_down(cls, event, group = all):
        """
        Processes any KEYDOWN events for buttons which require these.
        """
        cls.input_processed = False
        group_list = cls.get_group(group)
        #If any button in the current scope requires keyboard inputs / has focus:
        if cls.input_lock in group_list:
            if "Key_down" in cls.input_lock.actions:
                cls.input_lock.Key_down(event)
        return


    @classmethod
    def Set_cursor_pos(cls, pos, group = all):
        """
        Updates the cursor position. Required for e.g. sliders.
        """
        group_list = cls.get_group(group)
        if not cls.input_lock:
            for button in group_list:
                if "Set_cursor_pos" in button.actions:
                    button.Set_cursor_pos(pos)
        elif cls.input_lock in group_list:
            if "Set_cursor_pos" in cls.input_lock.actions:
                cls.input_lock.Set_cursor_pos(pos)


    @classmethod
    def convert_scroll(cls, value):
        """
        Converts a mouse button value into a scroll value.
        Scrolling down gives negative scroll values.
        """
        if value < 4:
            return 0
        elif value % 2:
            return round((value - 2) / 2) * cls.scroll_factor
        else:
            return round(-(value - 3) / 2) * cls.scroll_factor


    @classmethod
    def Scale(cls, scale, group = all, relative_scale = True, *, center = (0, 0), px_center = None):
        """
        Scales all buttons in the given group by / to a certain scaling factor.

        relative_scale: bool - Determines whether the given scale value is absolute (Button.scale = val), or is a scaling factor relative to their current scale.
        center: tuple - The absolute coordinates around which the scaling should take place.
        px_center: Tuple - The display coordinates around which the scaling should take place. If these are passed in, the 'center' coordinates are ignored.
        """
        if not isinstance(scale, (float, int)):
            raise TypeError(f"scale must be type 'int' or 'float', not type '{type(scale).__name__}'")
        elif scale == 0:
            raise ValueError(f"Cannot scale buttons to scale '0'")

        if px_center:
            px_center = cls.Verify_iterable(px_center, 2)
        else:
            center = cls.Verify_iterable(center, 2)

        for button in cls.get_group(group):
            if not relative_scale:
                scale_factor = cls.Clamp(scale, button.min_scale, button.max_scale) / button.scale
            else:
                scale_factor = cls.Clamp(scale, button.min_scale / button.scale, button.max_scale / button.scale)
            if px_center: #Transform the pixel coordinates to raw coordinates
                center = tuple(i / button.scale for i in px_center)

            #Apply the translation to make sure the given coordinates stay at the same place
            button._move(tuple(i * (1 / scale_factor - 1) for i in center))

            button.scale *= scale_factor

    @classmethod
    def Move(cls, offset, group = all, scale = False):
        """
        Moves all buttons in the given group by a certain offset.

        offset: int / float / tuple - An number, or iterable containing two numbers, for how much the Buttons in the group should be moved. If a number is given, this movement is applied in both directions.
        group: * - The group to which the translation should be applied.
        scale: bool - Determines whether the given values should be scaled to the Buttons' scale before they are applied.
        """
        if hasattr(offset, "__iter__"):
            cls.Verify_iterable(offset, 2, (int, float))
        else:
            offset = (offset, offset)
        #Set the "button_offset" variable to prevent scaling to affect other buttons too.
        b_offset = offset

        for button in cls.get_group(group):
            if scale:
                b_offset = button.scaled(offset, False)
            button._move(b_offset)


    @classmethod
    def Clear(cls, group = all):
        """
        Clears all user inputs from the buttons in the given group.

        group: * - The group of Buttons which should be cleared.
        """
        for button in cls.get_group(group):
            button.Clear()

    @classmethod
    def Draw(cls, screen, group = all):
        """
        Draw all buttons in the specified group to the screen / Surface provided.
        """
        #Select the correct button group
        group_list = cls.get_group(group)
        #Click all buttons without the "Cursor Lock".
        for button in reversed(group_list):
            if button is not cls.input_lock:
                button.Draw(screen)
        #Draw the button with the "Input Lock" last, to make it always appear on top.
        if cls.input_lock:
            if cls.input_lock in group_list:
                cls.input_lock.Draw(screen)


    @classmethod
    def Align(cls, rect, limit_size, pos):
        """
        Creates an aligned Rect, with the given size when placed onto a surface with size 'limit_size'.
        Functionally just a wrapper around AlignX and AlignY.

        rect: A pygame.Rect object or size tuple to be aligned.
        limit_size: A pygame.Rect object or size tuple to which the rect should be aligned.
        pos: The (text) containing the information on how the rect should be aligned (e.g. "top", "midleft", etc.). Defaults to centered if no alignment info is given for an axis, or the given info is invalid.
        """
        if isinstance(limit_size, pygame.Rect):
            #If the provided 'limit_size' is a Rect, extract its size
            limit_size = limit_size.size
        pos = pos.lower()

        rect = cls.AlignX(rect, limit_size[0], pos)
        rect = cls.AlignY(rect, limit_size[1], pos)

        return rect

    @staticmethod
    def AlignX(rect, max, pos):
        """
        Creates a rectangle that is correctly aligned horizontally.

        rect: A pygame.Rect object or size tuple to be aligned.
        limit_size: A pygame.Rect object or size tuple to which the rect should be aligned.
        pos: The (text) containing the information on how the rect should be aligned (e.g. "top", "midleft", etc.). Defaults to centered if no alignment info is given for an axis, or the given info is invalid.
        """
        if isinstance(rect, pygame.Rect):
            #Make sure the provided Rect is aligned correctly at the start
            rect.left = 0
        elif isinstance(rect, int):
            #If only an int (width) is given, create a zero height corresponding pygame.Rect
            rect = pygame.Rect((0, 0), (rect, 0))
        else:
            #If the provided 'rect' argument is not a rect, but just a tuple containing a size, create a new Rect
            rect = pygame.Rect((0, 0), rect)
        if isinstance(max, pygame.Rect):
            max = max.width
        pos = pos.lower()

        if "left" in pos:
            pass
        elif "right" in pos:
            rect.right = max
        else:
            rect.centerx = math.floor(max / 2)
        return rect

    @staticmethod
    def AlignY(rect, max, pos):
        """
        Creates a rectangle that is correctly aligned vertically.

        rect: A pygame.Rect object or size tuple to be aligned.
        limit_size: A pygame.Rect object or size tuple to which the rect should be aligned.
        pos: The (text) containing the information on how the rect should be aligned (e.g. "top", "midleft", etc.). Defaults to centered if no alignment info is given for an axis, or the given info is invalid.
        """
        if isinstance(rect, pygame.Rect):
            #Make sure the provided Rect is aligned correctly at the start
            rect.top = 0
        elif isinstance(rect, int):
            #If only an int (height) is given, create a zero width corresponding pygame.Rect
            rect = pygame.Rect((0, 0), (0, rect))
        else:
            #If the provided 'rect' argument is not a rect, but just a tuple containing a size, create a new Rect
            rect = pygame.Rect((0, 0), rect)
        if isinstance(max, pygame.Rect):
            max = max.height
        pos = pos.lower()

        if "top" in pos:
            pass
        elif "bottom" in pos:
            rect.right = max
        else:
            rect.centery = math.floor(max / 2)

        return rect


    def contains(self, position):
        """
        Tests whether a position is within the current (main) button.
        """
        #Test whether the pos input is valid
        position = self.Verify_iterable(position, 2)
        #If the position is within the corners. Note: Top and left have <=, whereas botom and right have < checks.
        #This is because the bottom / right values are actually just outside of the boxs' actual position
        if self.scaled(self.left) <= position[0] < self.scaled(self.right) and self.scaled(self.top) <= position[1] < self.scaled(self.bottom):
            return True
        else:
            return False

    @staticmethod
    def is_within(position, topleft, bottomright):
        """
        Tests whether a position is within two other corners. Basically a more generalised version of *.contains.
        """
        if topleft[0] <= position[0] < bottomright[0] and topleft[1] <= position[1] < bottomright[1]:
            return True
        else:
            return False


    @classmethod
    def Clamp(cls, value, minimum, maximum):
        """
        Returns a value which is as close to value as possible, but is minimum <= value <= maximum.
        """
        if maximum < minimum:
            raise ValueError(f"Maximum must be >= Minimum {maximum} and {minimum}")
        if hasattr(value, "__iter__"):
            return tuple(cls.Clamp(i, minimum, maximum) for i in value)
        return max(minimum, min(value, maximum))


    def Make_background_surface(self, inp, custom_size = None, scale_custom = False):
        """
        Makes a solid fill background if a colour was provided. If a surface was provided, returns that instead.
        If custom_size is set, will use that size instead of self.true_size.
        If scale_custom is True, and a custom size is given, that custom size will be scaled first.
        """
        if not custom_size:
            size = self.true_size
        elif scale_custom:
            size = self.scaled(custom_size)
        else:
            size = custom_size
        width, height = size
        #Set the background surface for the button. If one is provided, use
        # that one. Otherwise, make a new one with a solid color as given.
        if type(inp) == pygame.Surface:
            return pygame.transform.scale(inp, size)
        elif inp is None:
            return pygame.Surface(size, pygame.SRCALPHA)
        elif hasattr(inp, "__call__"):
            return inp()
        elif hasattr(inp[0], "__call__"): #If it is a tuple/list iterable with a function as its first item
            return inp[0](*(arg if arg != "*self*" else self for arg in inp[1:]))
        else:
            if type(self.style) is int:
                corner_radius = max(0, self.scaled(self.style))
            elif self.style.lower() == "square":
                corner_radius = 0
            elif self.style.lower() == "round":
                corner_radius = min(size)
            elif self.style.lower() == "smooth":
                corner_radius = self.scaled(12)
            else:
                raise ValueError(f"Invalid style value {self.style}")

            surface = pygame.Surface(size, pygame.SRCALPHA)
            pygame.draw.rect(surface, inp, ((0, 0), size), border_radius = corner_radius)
            return surface


    def Draw_border(self, surface, colour, border_width = 1, border_offset = 0, custom_size = None):
        """
        Draws a border around a surface.
        """
        border_offset = self.scaled(border_offset)
        border_width = self.scaled(border_width)

        if not border_width: #If after scaling, the border width is 0, don't try to draw anything, as doing so would colour the entire button.
            return

        style = self.style
        if custom_size:
            size = custom_size
        else:
            size = self.true_size
        if type(style) is int:
            corner_radius = max(0, self.scaled(style) - border_offset)
        elif style.lower() == "square":
            corner_radius = 0
        elif style.lower() == "round":
            corner_radius = min(size)
        elif style.lower() == "smooth":
            corner_radius = max(0, self.scaled(12) - border_offset)

        pygame.draw.rect(surface, colour, (2*(border_offset,), self.offset(size, 2*(border_offset,), (-2, -2))), border_width, corner_radius)


    @staticmethod
    def Verify_iterable(value, length = 2, data_types = None):
        """
        A function that verifies whether a given iterable has the required length, and whether all items in the iterable are of the correct types.
        """
        if not hasattr(value, "__iter__"):
            raise ValueError("Given value is not iterable")
        value_iterator = value.__iter__()
        #Get the first {length} items from the iterator.
        try:
            output = tuple(next(value_iterator) for _ in range(length))
        #If the iterator doesn't contain enough items, raise a ValueError
        except RuntimeError:
            raise ValueError("Given iterable contains too few items")
        if isinstance(data_types, (type, type(None))):
            data_types = [data_types]
        #If data_types == None,    or    all ites are of an allowed data_type: everything is fine; Else, raise an error.
        if not (data_types[0] is None    or    all(type(item) in data_types for item in output)):
            raise TypeError(f"Incorrect data type for items in iterable")
        #Test if the iterator did not contain more items:
        try:
            next(value_iterator)
        #If a StopIteration error is raised, this means the iterator contained
        #only two items, and thus was the correct size. In that case, return it.
        except StopIteration:
            return output
        else: #Otherwise, the iterator was too long. Raise an error.
            raise ValueError("Given iterable contains too many items")


    @classmethod
    def Verify_colour(cls, value):
        """
        Verifies whether a colour is in the correct format, and within the right range of values.
        """
        value = cls.Verify_iterable(value, 3, int)
        if all(0 <= i <= 255 for i in value):
            return value
        else:
            raise ValueError("All RGB values must be integers between 0 and 255")


    @classmethod
    def Verify_border(cls, border):
        """
        Verifies whether a border is in the correct format, and contains valid values.
        """
        if border:
            cls.Verify_iterable(border, 3)
            border_colour = cls.Verify_colour(border[0])
            if not all(isinstance(i, (int, float)) for i in border[1:]):
                raise TypeError("Border width and Border offset must be type 'int' or 'float'")
            return border_colour, border[1], border[2]
        else:
            return None


    @classmethod
    def Verify_background(cls, background):
        """
        Verifies whether a background is of a correct format / contains valid values.
        """
        if isinstance(background, pygame.Surface): #Pre-existing surface
            return background
        elif not background: #Empty background
            return None
        elif hasattr(background, "__call__"): #If it is itself a function
            return background
        elif isinstance(background, (list, tuple)) and background and hasattr(background[0], "__call__"): #If it is a tuple / list with a function as its first item.
            return background
        else:
            cls.Verify_colour(background)
            return background


    def Verify_functions(cls, functions):
        if not isinstance(functions, dict):
            raise TypeError(f"'functions' must be type 'dict', not type '{type(functions).__name__}'")
        if not all(isinstance(key, str) for key in functions):
            raise TypeError(f"All keys in 'functions' must be type 'str'")
        functions = {key.title(): value for key, value in functions.items()}
        return functions


    def Force_update(self):
        """
        A function that forces a button to get updated. Can be used when an attribute is changed which does not directly cause it to update.
        """
        #Set the updated parameter
        self.updated = True
        #Draw the button to make sure the button surface is updated too.
        self.Draw(pygame.Surface((1,1)))


    @staticmethod
    def offset(pos, offset_vector, scalar_vector = (1, 1)):
        """
        Returns a position with a certain offset.
        Also allows the offset to be multiplied by a scalar vector.
        """
        return tuple(pos[i] + offset_vector[i] * scalar_vector[i] for i in range(len(pos)))


    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, value):
        self.__scale = value
        self.updated = True
        for child in self.children:
            child.scale = value


    def _move(self, value):
        self.left += value[0]
        self.top += value[1]


    @property
    def font(self):
        #If the size of the font has changed, rebuild the font.
        if round(self.scale * self.font_size) != self.__font.get_height():
            self.__Make_font()
        #Return the font object.
        return self.__font


    @property
    def font_name(self):
        return self.__font_name

    @font_name.setter
    def font_name(self, value):
        self.__font_name = value
        self.__Make_font()
        for child in self.children:
            child.font_name = value


    @property
    def font_size(self):
        return self.__font_size

    @font_size.setter
    def font_size(self, value):
        self.__font_size = value
        self.updated = True
        for child in self.children:
            child.font_size = value


    def __Make_font(self):
        """
        Re-builds a font object based on self.font_name and self.font_size, as well as the current self.scale.
        """
        #pygame.font.Font is used in favor of pygame.font.SysFont, as SysFont's font sizes are inconsistent with the value given for the font.
        try:
            self.__font = pygame.font.Font(self.font_name, round(self.scale * self.font_size))
        except FileNotFoundError:
            font = pygame.font.match_font(self.font_name)
            if font is None: #If no matching font was found
                raise FileNotFoundError(f"No such font: '{self.font_name}'")
            self.__font = pygame.font.Font(font, round(self.scale * self.font_size))


    def _Call(self, action):
        """
        Calls a function, if it exists, for the action specified
        """
        if not action in self.functions: #If no function was specified for this action, ignore the fact that this function was called anyway
            return
        if isinstance(self.functions[action], (tuple, list)):
            self.functions[action][0](*(arg if arg != "*self*" else self for arg in self.functions[action][1:]))
        else:
            self.functions[action]()
        return


    @property
    def functions(self):
        return self.__functions
    @functions.setter
    def functions(self, value):
        self.__functions = self.Verify_functions(value)


    @property
    def text_colour(self):
        return self.__text_colour

    @text_colour.setter
    def text_colour(self, value):
        self.__text_colour = self.Verify_colour(value)
        self.updated = True
        for child in self.children:
            child.text_colour = value


    #Properties for all main positions of the button, much like a pygame.rect
    #Positions are not scaled by default. Run button.scaled() on the values to scale them if necessary
    @property
    def center(self):
        return (self.centerx, self.centery)
    @property
    def midbottom(self):
        return (self.centerx, self.bottom)
    @property
    def midtop(self):
        return (self.centerx, self.top)
    @property
    def midleft(self):
        return (self.left, self.centery)
    @property
    def midright(self):
        return (self.right, self.centery)
    @property
    def bottomleft(self):
        return (self.left, self.bottom)
    @property
    def bottomright(self):
        return (self.right, self.bottom)
    @property
    def topleft(self):
        return (self.left, self.top)
    @property
    def topright(self):
        return (self.right, self.top)
    @property
    def size(self):
        return (self.width, self.height)
    @property
    def bottom(self):
        return self.top + self.height
    @property
    def top(self):
        return self.__top
    @property
    def left(self):
        return self.__left
    @property
    def right(self):
        return self.left + self.width
    @property
    def height(self):
        return self.__height
    @property
    def width(self):
        return self.__width
    @property
    def centerx(self):
        return self.left + self.width / 2
    @property
    def centery(self):
        return self.top + self.height / 2

    @property
    def middle(self):
        """
        The middle of the button, pre-scaled.
        """
        return tuple(round(i / 2) for i in self.true_size)


    #Setter for all main positions of the button, much like a pygame.rect
    @center.setter
    def center(self, value):
        value = self.Verify_iterable(value, 2)
        self.centerx, self.centery = value
    @midbottom.setter
    def midbottom(self, value):
        value = self.Verify_iterable(value, 2)
        self.centerx, self.bottom = value
    @midtop.setter
    def midtop(self, value):
        value = self.Verify_iterable(value, 2)
        self.centerx, self.top = value
    @midleft.setter
    def midleft(self, value):
        value = self.Verify_iterable(value, 2)
        self.left, self.centery = value
    @midright.setter
    def midright(self, value):
        value = self.Verify_iterable(value, 2)
        self.right, self.centery = value
    @bottomleft.setter
    def bottomleft(self, value):
        value = self.Verify_iterable(value, 2)
        self.left, self.bottom = value
    @bottomright.setter
    def bottomright(self, value):
        value = self.Verify_iterable(value, 2)
        self.right, self.bottom = value
    @topleft.setter
    def topleft(self, value):
        value = self.Verify_iterable(value, 2)
        self.left, self.top = value
    @topright.setter
    def topright(self, value):
        value = self.Verify_iterable(value, 2)
        self.right, self.top = value
    @size.setter
    def size(self, value):
        value = self.Verify_iterable(value, 2)
        self.width, self.height = value
    @bottom.setter
    def bottom(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'bottom' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.top = value - self.height
    @top.setter
    def top(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'top' must by type 'int' or 'float', not type '{type(value).__name__}'")
        try:
            for child in self.children:
                child._move((0, value - self.top)) #Move children along with the main Button
        except AttributeError: pass #Catch error raised when .top is first set in __init__
        self.__top = value
        self.updated = True #Update button since moving might cause true_size to change
    @left.setter
    def left(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'left' must by type 'int' or 'float', not type '{type(value).__name__}'")
        try:
            for child in self.children:
                child._move((value - self.left, 0)) #Move children along with the main Button
        except AttributeError: pass #Catch error raised when .left is first set in __init__
        self.__left = value
        self.updated = True#Update button since moving might cause true_size to change
    @right.setter
    def right(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'right' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.left = value - self.width
    @centerx.setter
    def centerx(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'centerx' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.left = value - self.width / 2
    @centery.setter
    def centery(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'centery' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.top = value - self.height / 2
    @width.setter
    def width(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'width' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.__width = value
        self.updated = True
    @height.setter
    def height(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f"'height' must by type 'int' or 'float', not type '{type(value).__name__}'")
        self.__height = value
        self.updated = True

    #"True" size properties are used to prevent artifacting issues during scaling.
    # Although often the same as self.scaled(self.size), sometimes these will differ by a pixel due to rounding.
    @property
    def true_width(self):
        return self.scaled(self.right) - self.scaled(self.left)
    @property
    def true_height(self):
        return self.scaled(self.bottom) - self.scaled(self.top)
    @property
    def true_size(self):
        return (self.true_width, self.true_height)


    def get_rect(self):
        """
        Returns a pygame.Rect object of the unscaled button rectangle.
        """
        return pygame.Rect(self.topleft, self.size)

    def get_scaled_rect(self):
        """
        Returns a pygame.Rect object of the scaled button rectangle.
        """
        return pygame.Rect(self.scaled(self.topleft), self.true_size)



    def scaled(self, value, rounding = True):
        """
        Returns the scaled version of a value, or tuple of values.
        """
        if isinstance(value, (list, tuple)):
            return tuple(self.scaled(i, rounding) for i in value) #Recursion is amazing!
        elif isinstance(value, (float, int)):
            if rounding:
                return round(value * self.scale)
            else:
                return value * self.scale
        else:
            raise TypeError(f"Cannot scale type '{type(value).__name__}'.")

    def relative(self, pos):
        return self.offset(pos, self.scaled(self.topleft, False), (-1, -1))


#Add an attribute to this class that references itself, so other subclasses can easily acces the parent class object
#This is useful for allowing communications between subclasses
Buttons.Buttons = Buttons



from .Base import Buttons
from .Button  import Button
from .Slider import Slider

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame


class DropdownBox(Buttons):
    """
    Creates a DropdownBox, in which a user can select an input from a list of options.

    pos: (left, top) - The topleft position before scaling.
    size: (width, height) - The size before scaling.
    options: list, tuple - A list containing the values of all options to be added to the DropdownBox.
    hint: str - The text that is displayed in the main box when no option is selected.
    option_align: The alignment of the text on the dropped down 'option' buttons.
    value_align: The alignment of the current value on the main Button.
    hint_align: The alignment of the hint on the main Button.
    display_length: int - 0: unlimited (show all items regardless of how many there are).
                          +: Show up to n items, but if there are less, limit the length to the amount of items.
                          -: Show n items. If there are less items, show an empty area at the bottom of the list.
    button_spacing: int, (x, y) - The spacing between the buttons inside the Dropdown Box.
    style: "Square", "Round", int - Defines the radius of curvature of the buttons' corners.
    font_name: str - The name of the font that should be used for the DropdownBox.
    font_size: int - The size (in px) of the text.
    text_colour: (R, G, B) - The colour of the text in the DropdownBox.
    scroll_bar: None, int, Slider - The type of scrollbar to be included. Default styles 1 and 2 are available.
    background: pygame.Surface, (R, G, B), None, function - The background of the button if it is not selected.
    border: ((R, G, B), width, offset), None - The border that appears around the buttons in the DropdownBox.
    accent_background: pygame.Surface, (R, G, B), None, function - The background of the button if it is_selected. If set to None, will be the same as normal background.
    dropdown_background: pygame.Surface, (R, G, B), None, function - The background that is rendered behind the buttons on the dropdown section of the DropdownBox.
    functions: dict - Contains functions that should be called when a specific event occurs. The values should either be {"Click": func,} to call a function without arguments, or {"Click": (func, arg1, arg2, ...)} to call a function with arguments. If the Button itself is to be passed in as an argument, that argument can be passed in as '*self*'. This argument will automatically replaced when the function is actually called.
                    - "Select": Called whenever the DropdownBox is selected (dropped down).
                    - "Deselect": Called whenever the DropdownBox is deselected.
                    - "Update": Called whenever the state is changed.
                    - "Move": Called whenever the dropdown area is scrolled.
    groups: None, [___, ___] - A list of all groups to which a button is to be added.
    root: None, Button - The Button that is considered the 'root element' for this Button. Any function calls that need to include a 'self' Button, will include this root Button instead.
    independent: bool - Determines whether or not the button is allowed to set the input_lock, and is added to buttons.list_all. Mostly important for buttons which are part of another button.


    Inputs:
    *.state: int - Sets the index of the currently selected option. Set negative to deselect all options
    *.Add_option(*) - Adds an option to the list of possible options. See help(*.Add_option) for more information.
    *.Del_option(*) - Removes an option from the list of possible options. See help(*.Del_option) for more information.

    Outputs:
    *.value: * - The value of the currently selected item. Type can be wathever was given as the value.
    *.state: int - The index of the currently selected option. Is -1 if no option is selected.
    *.new_state: bool - Whether the DropdownBox has been set to a new state since the last time this variable was checked. Automatically resets once it is querried.

    *.is_selected: bool - Whether this DropdownBox object is selected at this point in time. I.E. Whether DropdownBox is expanded.
    """
    actions = ["LMB_down", "LMB_up", "Set_cursor_pos", "Scroll"]
    def __init__(self, pos, size,
                 options = [],
                 hint = "",
                 option_align = "center",
                 value_align = "left",
                 hint_align = "left", #Align left for consistency with text boxes
                 display_length = 0,
                 button_spacing = (0, 0),
                 style = "Square",
                 font_name = pygame.font.get_default_font(),
                 font_size = 22,
                 text_colour = (0, 0, 0),
                 hint_colour = (128, 128, 128),
                 scroll_bar = None,
                 background = (255, 255, 255),
                 border = ((63, 63, 63), 1, 0),
                 accent_background = (220, 220, 220),
                 dropdown_background = None,
                 functions = {},
                 group = None,
                 root = None,
                 independent = False
                 ):
        """
        Create a DropdownBox Button object. See help(type(self)) for more detailed information.
        """
        super().__init__(pos, size, font_name, font_size, group, root, independent)
        self.option_align = option_align
        self.hint = hint
        self.hint_align = hint_align
        self.value_align = value_align
        #Storing information required for later child buttons
        self.bg = self.Verify_background(background)
        self.accent_bg = self.Verify_background(accent_background)
        self.border = border
        self.style = style

        self.options = []
        self.button_list = []
        self.__state = -1
        self.__scrolled = 0
        self.new_state = False
        self.moved = True

        if type(button_spacing) in (int, float):
            self.spacing = (button_spacing, button_spacing)
        elif hasattr(button_spacing, "__iter__"):
            self.spacing = self.Verify_iterable(button_spacing, 2)
        else:
            raise ValueError("Incorrect spacing type")

        self.dropdown_bg = self.Verify_background(dropdown_background)
        self.display_length = display_length
        #Create the arrow button
        self.arrow = Button((self.right - self.height, self.top), (self.height, self.height), border = border, background = (Arrow_bg, "*self*", self.bg, self.accent_bg), accent_background = None, style = style, mode = "Toggle", root = self.root, independent = True)
        self.children.append(self.arrow)

        #Make the button containing the information about the currently selected option.
        self.main_button = Button(self.topleft, (self.width - self.height - self.spacing[0], self.height), font_name = font_name, font_size = font_size, border = border, background = background, accent_background = accent_background, style = style, root = self.root, independent = True)
        self.children.append(self.main_button)

        if scroll_bar:
            self.scroll_bar = Make_scroll_bar(self, scroll_bar)
            self.scroll_bar.Set_slider_primary(self.scroll_bar.height)
            self.children.append(self.scroll_bar)
        else:
            self.scroll_bar = None

        self.text_colour = self.Verify_colour(text_colour)
        self.hint_colour = self.Verify_colour(hint_colour)

        self.functions = functions
        #Add in all the options
        for option in options:
            self.Add_option(option)
        self.Draw(pygame.Surface((1, 1))) #Makes sure all attributes are set-up correctly


    def LMB_down(self, pos):
        #Test if any of the two header boxes was clicked:
        if self.main_button.contains(pos):
            self.Claim_input()
            self.is_selected = not self.is_selected
            self.updated = True
        elif self.arrow.contains(pos):
            self.Claim_input()
            self.is_selected = not self.is_selected
            self.updated = True
        #Test if the scroll_bar (if any) contained the location
        elif self.scroll_bar and self.scroll_bar.contains(pos):
            self.scroll_bar.LMB_down(pos)
            self.Set_lock()
        #Test if the any of the buttons in the dropdown space contain the clicking point
        elif self.is_selected and self.is_within(pos, (self.scaled(self.left), self.scaled(self.bottom) + self.scaled(self.spacing[1])), (self.scaled(self.right), self.scaled(self.bottom) + self.scaled(self.spacing[1]) + self._true_pixel_length)):
            #Claim the input. Even if no button is "hit", it was within the surface of the dropdown box
            self.Claim_input()
            #Calculate the new relative pos, when it is taken relative to the dropdown spaces' coordinates
            for button_nr, button in enumerate(self.button_list):
                if button.contains((pos[0], pos[1] + self.scrolled_px)):
                    self.state = button_nr
                    self.is_selected = False
                    #Stop checking from here, as no buttons overlap, thus no other buttons will .contain(rel_pos)
                    return
        elif self.is_selected:
            self.is_selected = False

    def LMB_up(self, pos):
        if self.scroll_bar:
            self.scroll_bar.LMB_up(pos)

    def Set_cursor_pos(self, pos):
        if self.scroll_bar:
            self.scroll_bar.Set_cursor_pos(pos)


    def Scroll(self, value, pos):
        if self.arrow.value: #If self is selected / the menu is expanded downwards
            if self.contains(pos):
                pass
            elif self.is_within(pos, (self.scaled(self.left), self.scaled(self.bottom) + self.scaled(self.spacing[1])), (self.scaled(self.right), self.scaled(self.bottom) + self.scaled(self.spacing[1]) + self._true_pixel_length)): #If the position lies withing the expanded section, perform scrolling.
                self.scrolled_px += self.Buttons.scroll_factor * value
                self.Claim_input()


    def Scale(self, scale, relative_scale = True, *, center = (0, 0), px_center = None):
        super().Scale(scale, self, relative_scale, center = center, px_center = px_center)


    def Move(self, offset, scale = False):
        super().Move(offset, self, scale)


    def Clear(self):
        self.is_selected = False
        self.state = -1
        #Lock is automatically released in property setter


    def Draw(self, screen, pos = None):
        """
        Draw the button to the screen.
        """
        #Set the correct position for the dropdown_surface
        dropdown_pos = (self.scaled(self.left), self.scaled(self.bottom) + self.scaled(self.spacing[1]))
        if pos is not None:
            offset = self.offset(pos, self.scaled(self.topleft), (-1, -1))
            dropdown_pos = self.offset(dropdown_pos, offset)

        self.scrolled
        #If the box has been updated, re-draw this stuff:
        if self.updated:
            #Set self.moved to True to make sure the dropdown_surface also gets updated to the latest state
            self.moved = True
            #Draw the main / header surface
            #Re-draw self.surface (containing the header and the arrow)
            self.surface = self.Make_background_surface(None)
            self.main_button.Draw(self.surface, (0, 0))
            self.arrow.Draw(self.surface, (self.true_width - self.arrow.true_width, 0))

            #Re-build the button surface
            #Re-draw self.button_surface (the pre-rendered surface containing ALL buttons stacked underneath each other)
            self.button_surface = pygame.Surface((self.true_width - (self.scroll_bar.true_width if self.scroll_bar else 0), self.scaled(self.button_list[-1].bottom) - self.scaled(self.button_list[0].top) if self.button_list else 0), pygame.SRCALPHA)
            for button in self.button_list:
                button.Draw(self.button_surface, (0, button.scaled(button.top) - self.scaled(self.button_list[0].top)))

            self.updated = False

        if self.moved:
            #re-draw self.dropdown_surface (The cut-to-size version of self.button_surface), including the potential scroll_bar
            self.dropdown_surface = self.Make_background_surface(self.dropdown_bg, (self.true_width, self._true_pixel_length))
            self.dropdown_surface.blit(self.button_surface, (0, -self.scrolled_px))
            if self.scroll_bar:
                self.scroll_bar.Draw(self.dropdown_surface, (self.true_width - self.scroll_bar.true_width, 0))

            self.moved = False


        if self.is_selected:
            screen.blit(self.dropdown_surface, dropdown_pos)
        screen.blit(self.surface, pos or self.scaled(self.topleft))
        return


    def Add_option(self, value, sort = False, set_to = False):
        """
        Add an option to the dropdown lists' options.
        sort: bool - Whether the new value should be sorted between old values. If False, the value is simply appended to the end.
        set_to: bool - Whether the new value should be automatically switched to as the currently selected option.
        """
        #Save and clear the state, as it can change by inserting a new button in between
        state = self._state
        self._state = -1
        if sort:
            #The index is the first item, for which value sorts before the item
            index = min([ind for ind, option in enumerate(self.options) if value == sorted([value, option])[0]], default = len(self.options))
        else:
            index = len(self.options)
        self.options.insert(index, value)
        new_button = Button((self.left, self.bottom + self.spacing[1] + index * (self.height + self.spacing[1])),
                            (self.width - (self.scroll_bar.width + self.spacing[0] if self.scroll_bar else 0), self.height),
                            text = str(value),
                            text_align = self.option_align,
                            style = self.style,
                            background = self.bg,
                            accent_background = self.accent_bg,
                            text_colour = self.text_colour,
                            font_name = self.font_name,
                            font_size = self.font_size,
                            border = self.border,
                            root = self.root,
                            independent = True
                            )
        self.button_list.insert(index, new_button)
        self.children.append(new_button)
        new_button.scale = self.scale

        #Move all following buttons down (if necessary) to make space for the new button
        for button in self.button_list[index + 1:]:
            button.top += self.height + self.spacing[1]

        #Set self.state to the correct value again
        if set_to:
            self.state = index #Set the new state, including running ._Call
        elif index <= self._state: #If the new item is before the current one, shift the index by 1 as well, and don't run ._Call
            self._state = state + 1
        else: #Otherwise, set the state back to what it was already
            self._state = state

        #Update the scroll_bar size if present
        if self.scroll_bar:
            self.scroll_bar.height = self._display_pixel_length
            if self.display_length == 0:
                self.scroll_bar.Set_slider_primary(self.scroll_bar.height)
            elif self.display_length > 0:
                self.scroll_bar.Set_slider_primary(round(self.scroll_bar.height * min(len(self.options), self.display_length) / len(self.options)))
            else:
                self.scroll_bar.Set_slider_primary(round(self.scroll_bar.height * min(1, -self.display_length / len(self.options))))
            self.scroll_bar.slider.limits[3] = self.scroll_bar.bottom
        return


    def Del_option(self, option = None, index = None):
        """
        Allows an option to be removed from the dropdown list. The option can be indicated either by the options' value (option = value), or the options' index (index = value).
        If the provided option does not exist, the function will cancel the operation gracefully.

        Returns:
        True if deletion was successful.
        False if deletion was unsuccessful.
        """
        #Determine the index of the item to be removed, or cancel the operation if the given value is invalid
        if isinstance(index, int):
            if index >= len(self.options) or -index > len(self.options): #If the index is invalid (too big positive, or too small negative)
                return False
            index %= len(self.options)
        else:
            #Try to find the index of the specified item
            try:
                index = self.options.index(option)
            except ValueError: #If the option is not inside the list:
                return False

        #Save and clear the state, as it can change by removing a button
        state = self._state
        self._state = -1

        #Remove all references to the button / option from this butttons' lists
        self.children.remove(self.button_list[index])
        self.options.pop(index)
        self.button_list.pop(index)

        #Move the relevant buttons upwards again
        for button in self.button_list[index:]:
            button.top -= self.height + self.spacing[1]

        #Reduce the state by 1, if the selected button came after the current button
        if state > index:
            self._state = state - 1
        elif state < index:
            self._state = state
        else:
            #If state == index, don't change self.state again. In that case, the selected item was the one which was removed, so we shouldn't set a new item.
            #._Call should still be run though as the state did change.
            self.root._Call("Update")

        #Update the scroll_bar size if present
        if self.scroll_bar:
            self.scroll_bar.height = self._display_pixel_length
            if self.display_length == 0:
                self.scroll_bar.Set_slider_primary(self.scroll_bar.height)
            elif self.display_length > 0:
                self.scroll_bar.Set_slider_primary(round(self.scroll_bar.height * min(len(self.options), self.display_length) / max(1, len(self.options))))
            else:
                self.scroll_bar.Set_slider_primary(round(self.scroll_bar.height * min(1, -self.display_length / max(1, len(self.options)))))
            self.scroll_bar.slider.limits[3] = self.scroll_bar.bottom #Update the bottom limit of the scroll_bar slider

        return True


    @property
    def value(self):
        if self._state >= 0:
            return self.options[self._state]
        else:
            return None
    @value.setter
    def value(self, value):
        if value in self.options:
            self.state = self.options.index(value)
        elif value is None:
            self.state = -1
        else:
            raise ValueError("Value not in options list")

    #_state is for internal use only. It allows the Button to update the state, without accidentally triggering the ._Call.
    # In doing so, it allows any user-set values to still run ._Call, and perform all required checks with minimal code duplicates.
    @property
    def _state(self):
        return self.__state
    @_state.setter
    def _state(self, value):
        #Verify that the index is not too big to prevent getting into an unrecoverable state if the state is out of bounds.
        if value >= len(self.options):
            raise IndexError("Index out of range")
        #Clear the currently selected button (if any)
        if self.__state >= 0:
            self.button_list[self.__state].value = False
        #If the new value selects an actual button:
        if value >= 0:
            self.__state = value
            self.button_list[value].value = True
            self.main_button.text = self.button_list[value].text
            self.main_button.text_colour = self.text_colour
            self.main_button.text_align = self.value_align
        else:
            self.__state = -1
            self.main_button.text = self.hint
            self.main_button.text_colour = self.hint_colour
            self.main_button.text_align = self.hint_align
        self.new_state = True
        self.updated = True
        self.moved = True


    @property
    def state(self):
        return self.__state
    @state.setter
    def state(self, value):
        self._state = value
        self.root._Call("Update")

    @property
    def new_state(self):
        new_state = self.__new_state
        self.__new_state = False
        return new_state
    @new_state.setter
    def new_state(self, value):
        self.__new_state = True


    @property
    def scrolled(self):
        if self.scroll_bar and self.scroll_bar.moved:
            self.__scrolled = self.scroll_bar.value
            self.moved = True
        return self.__scrolled

    @scrolled.setter
    def scrolled(self, value):
        #Make sure the scrolled value cannot exceed the limits of the space in the box
        value = self.Clamp(value, 0, 1)
        self.moved = True

        self.__scrolled = value
        if self.scroll_bar:
            self.scroll_bar.value = value

    @property
    def scrolled_px(self):
        #Calculate the required height (in px)
        req_height = 0 if not self.options else self.scaled(self.button_list[-1].bottom) - self.scaled(self.button_list[0].top)
        #Calculate the required height change that has to be accomodated by scrolling (in px)
        max_scroll_height = max(0, req_height - self._true_pixel_length) #Get the total distance (in px) the surface must be scrolled

        return round(max_scroll_height * self.scrolled)

    @scrolled_px.setter
    def scrolled_px(self, value):
        #Calculate the required height (in px)
        req_height = 0 if not self.options else self.scaled(self.button_list[-1].bottom) - self.scaled(self.button_list[0].top)
        #Calculate the required height change that has to be accomodated by scrolling (in px)
        max_scroll_height = max(0, req_height - self._true_pixel_length) #Get the total distance (in px) the surface must be scrolled

        #Only update the scrolled value, if the DD can be scrolled
        if max_scroll_height:
            self.scrolled = value / max_scroll_height


    @property
    def is_selected(self):
        return self.arrow.value

    @is_selected.setter
    def is_selected(self, value):
        if value == self.arrow.value:
            if value:
                self.Set_lock()
            else:
                self.Release_lock(False)
            return
        elif value:
            self.arrow.value = True
            self.Set_lock()
            self.root._Call("Select")
        else:
            self.arrow.value = False
            self.scrolled = 0
            self.moved = True
            self.Release_lock(False)
            self.root._Call("Deselect")
        self.updated = True


    @property
    def _display_pixel_length(self):
        """
        The length of the dropdown_surface in pixels (before scaling).
        """
        if self.display_length == 0:
            return max(0, len(self.options) * (self.height + self.spacing[1]) - self.spacing[1])
        elif self.display_length > 0:
            return max(0, min(self.display_length, len(self.options)) * (self.height + self.spacing[1]) - self.spacing[1])
        else:
            return abs(self.display_length) * (self.height + self.spacing[1]) - self.spacing[1]

    @property
    def _true_pixel_length(self):
        """
        The true / scaled pixel length of the dropdown_surface.
        """
        if self.scroll_bar:
            return self.scroll_bar.true_height
        else:
            return self.scaled(self.bottom + self.spacing[1] + self._display_pixel_length) - self.scaled(self.bottom + self.spacing[1])

def Arrow_bg(self, bg, accent_bg):
    """
    The function that will create the background for the dropdown arrow button.
    For internal use only. This function is therefore also not imported by __init__.py
    """
    #Just leave the making of the buttons background to the default function. Not gonna bother re-doing that here (because why would I?)
    if self.value:
        surface = self.Make_background_surface(accent_bg)
    else:
        surface = self.Make_background_surface(bg)

    #Draw the arrow so characteristic to dropdown boxes
    if not self.value:
        arrow_coords = (
            (round(self.true_width * 1/6), round(self.true_height * 1/3)), #Top left
            (round(self.true_width * 1/2), round(self.true_height * 2/3)), #Bottom
            (round(self.true_width * 5/6), round(self.true_height * 1/3)), #Top right
            )
    else:
        arrow_coords = (
            (round(self.true_width * 1/6), round(self.true_height * 2/3)), #Bottom left
            (round(self.true_width * 1/2), round(self.true_height * 1/3)), #Top
            (round(self.true_width * 5/6), round(self.true_height * 2/3)), #Bottom right
            )
    pygame.draw.polygon(surface, self.border[0] if self.border else (63, 63, 63), arrow_coords)

    return surface


def Make_scroll_bar(self, scroll_bar):
    """
    Makes a scroll_bar for a DropdownBox.
    For internal use only. This function is therefore also not imported by __init__.py
    """
    if isinstance(scroll_bar, Slider):
        scroll_bar.right = self.right
        scroll_bar.height = self._display_pixel_length
        scroll_bar.top = self.bottom
        scroll_bar.root = self.root
        return scroll_bar
    if scroll_bar == 1:
        size = (15, self._display_pixel_length)
        pos = (self.right - size[0], self.bottom)
        style = "Round"
        background = None
        border = None
        slider_bg = (220, 220, 220)
        slider_accent_bg = (127, 127, 127)
        slider_border = None
        return Slider(pos, size, style = style, background = background, border = border, slider_background = slider_bg, slider_border = slider_border, orientation = 1, root = self.root, independent = True)
    elif scroll_bar == 2:
        size = (15, self._display_pixel_length)
        pos = (self.right - size[0], self.bottom)
        slider_feature_text = "|||"
        slider_feature_size = 9
        return Slider(pos, size, slider_feature_text = slider_feature_text, slider_feature_size = slider_feature_size, orientation = 1, root = self.root, independent = True)
    else:
        raise ValueError(f"Unsupported scroll_bar style: {repr(scroll_bar)}")



from .Base import Buttons
from .Button import Button

import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame


class Slider(Buttons):
    """
    Creates a Slider, which allows the user to input a value within a given range.

    pos: (left, top) - The topleft position before scaling.
    size: (width, height) - The size before scaling.
    value_range: (a, b) - The range between which values the slider should (linearly) interpolate.
    orientation: "auto", int - The orientation of the Slider. In case orientation == "auto", the longest direction will be seen as the primary direction. If orientation == 0, the Slider will be horizontal; if orientation == 1, the Slider will be vertical.
    style: "Square", "Round", int - Defines the radius of curvature of the buttons' corners.
    background: pygame.Surface, (R, G, B), None, function - The background of the button if it is not selected.
    border: ((R, G, B), width, offset), None - The border that appears around the Sliders' background.
    markings: int - The amount of markings to be drawn to the background. Set to 0 to disable all markings.
    edge_markings: bool - Whether or not markings should be present at the edges too, or should be spaced equally over the entire text_box. Can not be enabled when markings < 2.
    marking_colour: (R, G, B) - The colour the markings will have when drawn onto the Slider background.
    snap_radius: int, float - The radius (in pixels) in which the slider should snap towards any markings.
    slider_background: pygame.Surface, (R, G, B), None, function - The background of the slider if it is not selected.
    slider_border: ((R, G, B), width, offset), None - The border that appears around the slider.
    accent background: pygame.Surface, (R, G, B), None, function - The background of the slider if it is_selected. If set to None, will be the same as normal background.
    slider_feature_font: str - The name of the font that should be used for the slider feature.
    slider_feature_size: int - The size (in px) of the sliders' feature text.
    slider_feature_colour: (R, G, B) - The colour of the feature / text on the slider.
    slider_feature_align: The alignment of the feature text on the slider Button.
    slider_feature_text: str - The feature / text that will be rendered to the slider.
    slider_size: "auto", int, (width, height) - The size of the slider. If set to "auto", will automatically fit the slider to the direction orthogonal to the orientation.
    functions: dict - Contains functions that should be called when a specific event occurs. The values should either be {"Click": func,} to call a function without arguments, or {"Click": (func, arg1, arg2, ...)} to call a function with arguments. If the Button itself is to be passed in as an argument, that argument can be passed in as '*self*'. This argument will automatically replaced when the function is actually called.
                    - "Click": Called when the Slider is clicked.
                    - "Release": Called when the Slider is released.
                    - "Move": Called when the Slider is moved to a new location. Only called by user input.
    groups: None, [___, ___] - A list of all groups to which a button is to be added.
    root: None, Button - The Button that is considered the 'root element' for this Button. Any function calls that need to include a 'self' Button, will include this root Button instead.
    independent: bool - Determines whether or not the button is allowed to set the input_lock, and is added to buttons.list_all. Mostly important for buttons which are part of another button.

    Inputs:
    *.value: int, float - The current value of the Slider.
    *.Set_range: (a, b) - Sets the range between which values the slider should (linearly) interpolate. See help(*.Set_range) for more information.
    *.Set_slider_primary(*) - Sets the primary size of the slider. Mainly useful when the slider is used as a scrollbar. See help(*.Set_slider_primary) for more information.

    Outputs:
    *.value: float - The current value of the slider.
    *.moved: bool - Whether the slider has been moved since the last time this property has been checked. Automatically resets once it is querried.

    *.is_selected: bool - Whether this Slider object is selected at this point in time. I.E. Whether the user is currently moving the Slider.
    """
    actions = ["LMB_down", "LMB_up", "Set_cursor_pos"]
    def __init__(self, pos, size,
                 value_range = (0, 1),
                 start_value = 0,
                 orientation = "auto", #0 for horizontal, 1 for vertical
                 style = "Square",
                 #Background settings
                 background = (255, 255, 255), #Colour or pygame.Surface
                 border = ((63, 63, 63), 1, 0),
                 #Marking settings:
                 markings = 0, #The amount of markings that should be made. 0 to disable all markings.
                 edge_markings = False,
                 marking_colour = (127, 127, 127),
                 snap_radius = 0, #The distance for which the slider should snap to markings (in pixels).
                 #Settings for the slider button. Provide Button object for custom sliders.
                 slider_background = (220, 220, 220),
                 slider_border = ((63, 63, 63), 1, 0),
                 slider_accent_background = (191, 191, 191),
                 slider_feature_font = pygame.font.get_default_font(),
                 slider_feature_size = 22,
                 slider_feature_colour = (63, 63, 63),
                 slider_feature_align = "center",
                 slider_feature_text = "",
                 slider_size = "Auto",
                 #Other (miscelaneous) settings
                 functions = {},
                 group = None,
                 root = None,
                 independent = False,
                 ):
        """
        Create a Slider Button object. See help(type(self)) for more detailed information.
        """
        #Note: 'Slider' (captialised) in commments etc. refers to the main object (self)
        #      'slider' (non-capitalised) refers to the button that moves along the Slider (self.slider)
        super().__init__(pos, size, groups = group, root = root, independent = independent) #We don't care about the font, as this Button will not contain any text
        #Initialise the basic parameters of the Slider
        self.__value_range = self.Verify_iterable(value_range, 2) #Directly written to the private property, to prevent a chicken - egg problem with self.value
        if type(orientation) is int:
            self.orientation = orientation
        elif orientation.lower() == "auto":
            self.orientation = int(self.width < self.height)
        self.style = style
        #Set the background parameters for the Slider
        self.bg = self.Buttons.Verify_background(background)
        self.border = self.Verify_border(border)

        #Initialise any markers - Has to be done before making the slider object
        if markings == 1 and edge_markings:
            raise ValueError("Edge markings require at least 2 markings to be present")
        self.markings = markings
        self.edge_markings = edge_markings
        self.marking_colour = marking_colour

        #Create the sliding object (from now on referred to as "slider" (lower case))
        self.tmp_slider_size = slider_size
        self.slider = Make_slider(self, style, slider_size, slider_background, slider_accent_background, slider_border, markings, edge_markings, snap_radius, slider_feature_text, slider_feature_colour, slider_feature_align, slider_feature_font, slider_feature_size, self.orientation)
        self._value = self.start_value = start_value
        del self.tmp_slider_size
        self.children.append(self.slider)

        #Other
        self.is_selected = False
        self.moved = True #Indicates whether there is a chance the slider has moved. If so, the user can take action (if necessary).
        self.functions = functions
        self.Draw(pygame.Surface((1, 1))) #Makes sure all attributes are set-up correctly


    def LMB_down(self, pos):
        if self.slider.contains(pos):
            self.is_selected = True
            self.slider.LMB_down(pos)
            self.Set_lock()
            self.moved = True
            self._moved = True #Instruct the button that value is no longer equal to the value stored in __value
        elif self.contains(pos):
            self.is_selected = True
            #Move the slider to where we clicked (within limits of course)
            self.slider.LMB_down(self.slider.scaled(self.slider.center))
            self.Set_cursor_pos(pos)
            self.Set_lock()
        return

    def LMB_up(self, pos):
        if self.is_selected:
            self.slider.LMB_up(pos)
            self.is_selected = False
            self.moved = True #To make sure the button get's re-drawn correctly with the right background
            self._moved = True
        return

    def Set_cursor_pos(self, pos):
        if self.is_selected:
            slider_pos = self.slider.topleft
            self.slider.Set_cursor_pos(pos)
            if self.slider.topleft != slider_pos:
                self.moved = True
                self._moved = True


    def Scale(self, scale, relative_scale = True, *, center = (0, 0), px_center = None):
        super().Scale(scale, self, relative_scale, center = center, px_center = px_center)


    def Move(self, offset, scale = False):
        super().Move(offset, self, scale)


    def Clear(self):
        #Reset the slider position to its inital value
        self._value = self.start_value
        self.is_selected = False
        #Lock is automatically released in property setter


    def Draw(self, screen, pos = None):
        """
        Draw the button to the screen.
        """
        #Set draw positions for when a custom location is given
        if pos is not None:
            slider_pos = self.offset(self.slider.scaled(self.slider.topleft), self.offset(pos, self.scaled(self.topleft), (-1, -1)))
        else:
            pos = self.scaled(self.topleft)
            slider_pos = None

        #Update the button surface (if necessary)
        if self.updated:
            #Now, let's actually construct the surface
            self.surface = self.Make_background_surface(self.bg)
            if self.border:
                self.Draw_border(self.surface, *self.border)

            if self.markings:
                #Set up the information of the marking itself
                marking_height = self.rotated(self.true_size)[1] - 2*(self.scaled(self.border[1]) + self.scaled(self.border[2]) if self.border else 0)
                marking_width = self.scaled(1)
                marking_rect = pygame.Rect((0,0), self.rotated((marking_width, marking_height)))

                #Iterate over all markings, and draw them
                for coord in self.Marking_coords():
                    coord = self.scaled(coord)
                    marking_rect.center = self.rotated(coord, self.rotated(self.true_size)[1] / 2)
                    pygame.draw.rect(self.surface, self.marking_colour, marking_rect)

            self.updated = False

        screen.blit(self.surface, pos)
        self.slider.Draw(screen, slider_pos)
        return




    def rotated(self, value, other = None):
        """
        Returns a rotated version of a 2-item list / tuple, such that the primary dimension is always first in the tuple.
        In case the orientation is horizontal, it stays the same.
        In case the orientation is vertical, it becomes reversed.
        """
        if other is not None:
            value = (value, other)
        if self.orientation % 2:
            return tuple(reversed(value))
        else:
            return value

    def Marking_coords(self):
        try:
            slider_size = self.slider.size
        except AttributeError: #During setup, the coords are required to construct the slider object. Therefore, when slider does not exist, instead take the size from the temporary value.
            if type(self.tmp_slider_size) is int:
                slider_size = 2 * (self.tmp_slider_size,)
            elif type(self.tmp_slider_size) is not str:
                slider_size = self.tmp_slider_size
            elif self.tmp_slider_size.lower() == "auto":
                slider_size = 2 * (min(self.size),)
        coord_range = self.rotated(self.size)[0] - self.rotated(slider_size)[0] #The available pixels for the slider to move in
        offset = self.rotated(slider_size)[0] / 2 #The offset due to the slider having to fit within the sliders' width (to some degree)
        for i in range(self.markings):
            if self.edge_markings:
                yield(coord_range / (self.markings - 1) * i + offset)
            else:
                yield(coord_range / (self.markings + 1) * (i + 1) + offset)

    def Set_range(self, range, *args):
        """
        Set a new range for the slider. Can be done as either:
        set_range([min, max]) or
        set_range(min, max)
        """
        if args: #If the user passed the values in as two separate values, combine them into one tuple
            range = (range, args[0])
        self._value #Flush any _moved arguments, in case they haven't been processed yet.
        self.value_range = tuple(range) #Update the slider range
        self.value = self.Clamp(self._value, *sorted(self.value_range)) #Reset the value, to update the sliders' position

    def Set_slider_primary(self, value, limit_size = True):
        """
        Set the sliders' primary dimension / size (in the direction of travel).
        Set_slider_primary(value).
        If limit_size == True, the primary dimension cannot go below 1/2 the secondary dimension. This is to prevent the slider from disappearing completely if the primary is set too low.
        """
        if limit_size:
            #Make sure the slider primary can not accidentally be set too low.
            value = max(value, round(self.rotated(self.slider.size)[1] / 2))
        self._value #Flush any _moved arguments, in case they haven't been processed yet.
        if self.orientation % 2: #Update the sliders' size
            if self.slider.height != value:
                self.slider.height = value
        else:
            if self.slider.width != value:
                self.slider.width = value
        self._value = self._value #Reset the value to update the sliders' position
        #Update the sliders' snap points
        self.slider.snap = self.rotated(tuple(self.rotated(self.topleft)[0] - value / 2 + coord for coord in self.Marking_coords()), ()) + (self.slider.snap[2],)
        self.updated = True


    @property
    def is_selected(self):
        return self.__is_selected

    @is_selected.setter
    def is_selected(self, value):
        if value: #If the user selects the text box:
            self.__is_selected = True
            self.Set_lock()
        else: #If the user deselects the box:
            self.__is_selected = False
            self.Release_lock()

    @property
    def _value(self):
        #Note: No need to run ._Call() anywhere in this code. The fact that the .__value is only updated here isn't important. It was already called when the slider was actually moved.
        if self._moved: #If the user clicked on the slider, the value should be re-calculated from the slider position. If not, the value should be exactly what the user set.
            pos = round(self.rotated(self.offset(self.slider.topleft, self.topleft, (-1, -1)))[0])
            coord_range = self.rotated(self.size)[0] - self.rotated(self.slider.size)[0] #The available pixels for the slider to move in
            if coord_range == 0: #If the slider is the same size as the overall button, pre-emptively catch it to prevent a ZeroDivisionError
                self.__value = sum(self.value_range) / 2
            else:
                self.__value = self.Clamp(self.value_range[0] + pos / coord_range * (self.value_range[1] - self.value_range[0]), *sorted(self.value_range))
            self._moved = False
        return self.__value

    @_value.setter
    def _value(self, val):
        val = self.Clamp(val, *sorted(self.value_range))
        if self.value_range[0] == self.value_range[1]:
            self.slider.center = self.center
        else:
            coord_range = self.rotated(self.size)[0] - self.rotated(self.slider.size)[0] #The available pixels for the slider to move in
            self.slider.topleft = self.rotated(self.rotated(self.topleft)[0] + (val - self.value_range[0]) / (self.value_range[1] - self.value_range[0]) * coord_range, self.rotated(self.center)[1] - self.rotated(self.slider.size)[1] / 2)
        self.__value = val
        self._moved = False

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self.root._Call("Move")

    @property
    def value_range(self):
        return self.__value_range

    @value_range.setter
    def value_range(self, value):
        self.__value_range = self.Verify_iterable(value, 2)
        self._value += 0

    @property
    def moved(self):
        moved = self.__moved
        self.__moved = False
        return moved

    @moved.setter
    def moved(self, value):
        self.__moved = value


def Make_slider(self, style, size, background, accent_background, border, markings, edge_markings, snap_radius, feature_text, feature_colour, feature_align, feature_font, feature_size, orientation):
    """
    Make a slider Button.
    For internal use only. This function is therefore also not imported by __init__.py
    """
    if type(size) is int:
        size = 2 * (size,)
    elif type(size) is not str:
        pass
    elif size.lower() == "auto":
        #Primary direction: min(self.size)
        #Secondary direction: the Sliders' height in the secondary direction
        size = self.rotated((min(self.size), self.rotated(self.size)[1]))

    if self.orientation % 2:
        limits = (self.left - size[0], self.right + size[0], self.top, self.bottom)
    else:
        limits = (self.left, self.right, self.top - size[1], self.bottom + size[1])
    #pos is irrelevant, as it is set by the value setter anyway
    return Button((0, 0), size, mode = "Hold", style = style,
                background = background,
                text = feature_text,
                text_colour = feature_colour,
                text_align = feature_align,
                font_name = feature_font,
                font_size = feature_size,
                orientation = orientation,
                accent_background = accent_background,
                border = border,
                dragable = self.rotated((True, False)),
                limits = limits,
                snap = self.rotated(tuple(self.rotated(self.topleft)[0] - (self.rotated(size)[0] / 2) + coord for coord in self.Marking_coords()), ()) + (snap_radius,),
                root = self.root,
                independent = True,
                )
